import logging
import typing as tp

from httpcore import Request, Response

from hishel._headers import Vary, parse_cache_control

from ._utils import (
    BaseClock,
    Clock,
    extract_header_values,
    extract_header_values_decoded,
    generate_key,
    get_safe_url,
    header_presents,
    parse_date,
)

logger = logging.getLogger("hishel.controller")

HEURISTICALLY_CACHEABLE_STATUS_CODES = (200, 203, 204, 206, 300, 301, 308, 404, 405, 410, 414, 501)
HTTP_METHODS = ["GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE", "PATCH"]

__all__ = ("Controller", "HEURISTICALLY_CACHEABLE_STATUS_CODES")


def get_updated_headers(
    stored_response_headers: tp.List[tp.Tuple[bytes, bytes]],
    new_response_headers: tp.List[tp.Tuple[bytes, bytes]],
) -> tp.List[tp.Tuple[bytes, bytes]]:
    updated_headers = []

    checked = set()

    for key, value in stored_response_headers:
        if key not in checked and key.lower() != b"content-length":
            checked.add(key)
            values = extract_header_values(new_response_headers, key)

            if values:
                updated_headers.extend([(key, value) for value in values])
            else:
                values = extract_header_values(stored_response_headers, key)
                updated_headers.extend([(key, value) for value in values])

    for key, value in new_response_headers:
        if key not in checked and key.lower() != b"content-length":
            values = extract_header_values(new_response_headers, key)
            updated_headers.extend([(key, value) for value in values])

    return updated_headers


def get_freshness_lifetime(response: Response) -> tp.Optional[int]:
    response_cache_control = parse_cache_control(extract_header_values_decoded(response.headers, b"Cache-Control"))

    if response_cache_control.max_age is not None:
        return response_cache_control.max_age

    if header_presents(response.headers, b"expires"):
        expires = extract_header_values_decoded(response.headers, b"expires", single=True)[0]
        expires_timestamp = parse_date(expires)
        if expires_timestamp is None:
            return None
        date = extract_header_values_decoded(response.headers, b"date", single=True)[0]
        date_timestamp = parse_date(date)
        if date_timestamp is None:
            return None

        return expires_timestamp - date_timestamp
    return None


def get_heuristic_freshness(response: Response, clock: "BaseClock") -> int:
    last_modified = extract_header_values_decoded(response.headers, b"last-modified", single=True)

    if last_modified:
        last_modified_timestamp = parse_date(last_modified[0])
        if last_modified_timestamp is not None:
            now = clock.now()

            ONE_WEEK = 604_800

            return min(ONE_WEEK, int((now - last_modified_timestamp) * 0.1))

    ONE_DAY = 86_400
    return ONE_DAY


def get_age(response: Response, clock: "BaseClock") -> int:
    if not header_presents(response.headers, b"date"):
        # If the response does not have a date header, then it is impossible to calculate the age.
        # Instead of raising an exception, we return infinity to be sure that the response is not considered fresh.
        return float("inf")  # type: ignore

    date = parse_date(extract_header_values_decoded(response.headers, b"date")[0])
    if date is None:
        return float("inf")  # type: ignore

    now = clock.now()

    apparent_age = max(0, now - date)
    return int(apparent_age)


def allowed_stale(response: Response) -> bool:
    response_cache_control = parse_cache_control(extract_header_values_decoded(response.headers, b"Cache-Control"))

    if response_cache_control.no_cache:
        return False

    if response_cache_control.must_revalidate:
        return False

    return True


class Controller:
    def __init__(
        self,
        cacheable_methods: tp.Optional[tp.List[str]] = None,
        cacheable_status_codes: tp.Optional[tp.List[int]] = None,
        cache_private: bool = True,
        allow_heuristics: bool = False,
        clock: tp.Optional[BaseClock] = None,
        allow_stale: bool = False,
        always_revalidate: bool = False,
        force_cache: bool = False,
        key_generator: tp.Optional[tp.Callable[[Request, tp.Optional[bytes]], str]] = None,
    ):
        self._cacheable_methods = []

        if cacheable_methods is None:
            self._cacheable_methods.append("GET")
        else:
            for method in cacheable_methods:
                if method.upper() not in HTTP_METHODS:
                    raise RuntimeError(
                        f"Hishel does not support the HTTP method `{method}`.\n"
                        f"Please use the methods from this list: {HTTP_METHODS}"
                    )
                self._cacheable_methods.append(method.upper())

        self._cacheable_status_codes = cacheable_status_codes if cacheable_status_codes else [200, 301, 308]
        self._cache_private = cache_private
        self._clock = clock if clock else Clock()
        self._allow_heuristics = allow_heuristics
        self._allow_stale = allow_stale
        self._always_revalidate = always_revalidate
        self._force_cache = force_cache
        self._key_generator = key_generator or generate_key

    def is_cachable(self, request: Request, response: Response) -> bool:
        """
        Determines whether the response may be cached.

        The only thing this method does is determine whether the
        response associated with this request can be cached for later use.
        `https://www.rfc-editor.org/rfc/rfc9111.html#name-storing-responses-in-caches`
        lists the steps that this method simply follows.
        """
        method = request.method.decode("ascii")
        force_cache = request.extensions.get("force_cache", None)

        if response.status not in self._cacheable_status_codes:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    f"as not cachable since its status code ({response.status})"
                    " is not in the list of cacheable status codes."
                )
            )
            return False

        if response.status in (301, 308):
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as cachable since its status code is a permanent redirect."
                )
            )
            return True

        # the request method is understood by the cache
        if method not in self._cacheable_methods:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    f"as not cachable since the request method ({method}) is not in the list of cacheable methods."
                )
            )
            return False

        if force_cache if force_cache is not None else self._force_cache:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as cachable since the request is forced to use the cache."
                )
            )
            return True

        response_cache_control = parse_cache_control(extract_header_values_decoded(response.headers, b"cache-control"))
        request_cache_control = parse_cache_control(extract_header_values_decoded(request.headers, b"cache-control"))

        # the response status code is final
        if response.status // 100 == 1:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as not cachable since its status code is informational."
                )
            )
            return False

        # the no-store cache directive is not present (see Section 5.2.2.5)
        if request_cache_control.no_store:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as not cachable since the request contains the no-store directive."
                )
            )
            return False

        # note that the must-understand cache directive overrides
        # no-store in certain circumstances; see Section 5.2.2.3.
        if response_cache_control.no_store:
            if response_cache_control.must_understand:
                logger.debug(
                    (
                        f"Skipping the no-store directive for the resource located at {get_safe_url(request.url)} "
                        "since the response contains the must-understand directive."
                    )
                )
            else:
                logger.debug(
                    (
                        f"Considering the resource located at {get_safe_url(request.url)} "
                        "as not cachable since the response contains the no-store directive."
                    )
                )
                return False

        # a shared cache must not store a response with private directive
        # Note that we do not implement special handling for the qualified form,
        # which would only forbid storing specified headers.
        if not self._cache_private and response_cache_control.private:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as not cachable since the response contains the private directive."
                )
            )
            return False

        expires_presents = header_presents(response.headers, b"expires")
        # the response contains at least one of the following:
        # - a public response directive (see Section 5.2.2.9);
        # - a private response directive, if the cache is not shared (see Section 5.2.2.7);
        # - an Expires header field (see Section 5.3);
        # - a max-age response directive (see Section 5.2.2.1);
        # - if the cache is shared: an s-maxage response directive (see Section 5.2.2.10);
        # - a cache extension that allows it to be cached (see Section 5.2.3); or
        # - a status code that is defined as heuristically cacheable (see Section 4.2.2).
        if self._allow_heuristics and response.status in HEURISTICALLY_CACHEABLE_STATUS_CODES:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as cachable since its status code is heuristically cacheable."
                )
            )
            return True

        if not any(
            [
                response_cache_control.public,
                response_cache_control.private,
                expires_presents,
                response_cache_control.max_age is not None,
            ]
        ):
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as not cachable since it does not contain any of the required cache directives."
                )
            )
            return False

        logger.debug(
            (
                f"Considering the resource located at {get_safe_url(request.url)} "
                "as cachable since it meets the criteria for being stored in the cache."
            )
        )
        # response is a cachable!
        return True

    def _make_request_conditional(self, request: Request, response: Response) -> None:
        """
        Adds the precondition headers needed for response validation.

        This method will use the "Last-Modified" or "Etag" headers
        if they are provided in order to create precondition headers.

        See also (https://www.rfc-editor.org/rfc/rfc9111.html#name-sending-a-validation-reques)
        """

        if header_presents(response.headers, b"last-modified"):
            last_modified = extract_header_values(response.headers, b"last-modified", single=True)[0]
            logger.debug(
                (
                    f"Adding the 'If-Modified-Since' header with the value of '{last_modified.decode('ascii')}' "
                    f"to the request for the resource located at {get_safe_url(request.url)}."
                )
            )
        else:
            last_modified = None

        if header_presents(response.headers, b"etag"):
            etag = extract_header_values(response.headers, b"etag", single=True)[0]
            logger.debug(
                (
                    f"Adding the 'If-None-Match' header with the value of '{etag.decode('ascii')}' "
                    f"to the request for the resource located at {get_safe_url(request.url)}."
                )
            )
        else:
            etag = None

        precondition_headers: tp.List[tp.Tuple[bytes, bytes]] = []
        if last_modified:
            precondition_headers.append((b"If-Modified-Since", last_modified))
        if etag:
            precondition_headers.append((b"If-None-Match", etag))

        request.headers.extend(precondition_headers)

    def _validate_vary(self, request: Request, response: Response, original_request: Request) -> bool:
        """
        Determines whether the "vary" headers in the request and response headers are identical.

        See also (https://www.rfc-editor.org/rfc/rfc9111.html#name-calculating-cache-keys-with).
        """

        vary_headers = extract_header_values_decoded(response.headers, b"vary")
        vary = Vary.from_value(vary_values=vary_headers)
        for vary_header in vary._values:
            if vary_header == "*":
                return False  # pragma: no cover

            if extract_header_values(request.headers, vary_header) != extract_header_values(
                original_request.headers, vary_header
            ):
                return False

        return True

    def construct_response_from_cache(
        self, request: Request, response: Response, original_request: Request
    ) -> tp.Union[Response, Request, None]:
        """
        Specifies whether the response should be used, skipped, or validated by the cache.

        This method makes a decision regarding what to do with
        the stored response when it is retrieved from storage.
        It might be ready for use or it might need to be revalidated.
        This method mirrors the relevant section from RFC 9111,
        see (https://www.rfc-editor.org/rfc/rfc9111.html#name-constructing-responses-from).

        Returns:
            Response: This response is applicable to the request.
            Request: This response can be used for this request, but it must first be revalidated.
            None: It is not possible to use this response for this request.
        """

        # Use of responses with status codes 301 and 308 is always
        # legal as long as they don't adhere to any caching rules.
        if response.status in (301, 308):
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as valid for cache use since its status code is a permanent redirect."
                )
            )
            return response

        response_cache_control = parse_cache_control(extract_header_values_decoded(response.headers, b"Cache-Control"))
        request_cache_control = parse_cache_control(extract_header_values_decoded(request.headers, b"Cache-Control"))

        # request header fields nominated by the stored
        # response (if any) match those presented (see Section 4.1)
        if not self._validate_vary(request=request, response=response, original_request=original_request):
            # If the vary headers does not match, then do not use the response
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as invalid for cache use since the vary headers do not match."
                )
            )
            return None  # pragma: no cover

        # !!! this should be after the "vary" header validation.
        force_cache = request.extensions.get("force_cache", None)
        if force_cache if force_cache is not None else self._force_cache:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as valid for cache use since the request is forced to use the cache."
                )
            )
            return response

        # the stored response does not contain the
        # no-cache directive (Section 5.2.2.4), unless
        # it is successfully validated (Section 4.3)
        if (
            self._always_revalidate
            or response_cache_control.no_cache
            or response_cache_control.must_revalidate
            or request_cache_control.no_cache
        ):
            if self._always_revalidate:
                log_text = (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as needing revalidation since the cache is set to always revalidate."
                )
            elif response_cache_control.no_cache:
                log_text = (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as needing revalidation since the response contains the no-cache directive."
                )
            elif response_cache_control.must_revalidate:
                log_text = (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as needing revalidation since the response contains the must-revalidate directive."
                )
            elif request_cache_control.no_cache:
                log_text = (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as needing revalidation since the request contains the no-cache directive."
                )
            else:
                assert False, "Unreachable code "  # pragma: no cover
            logger.debug(log_text)
            self._make_request_conditional(request=request, response=response)
            return request

        freshness_lifetime = get_freshness_lifetime(response)

        if freshness_lifetime is None:
            logger.debug(
                (
                    "Could not determine the freshness lifetime of "
                    f"the resource located at {get_safe_url(request.url)}, "
                    "trying to use heuristics to calculate it."
                )
            )
            if self._allow_heuristics and response.status in HEURISTICALLY_CACHEABLE_STATUS_CODES:
                freshness_lifetime = get_heuristic_freshness(response=response, clock=self._clock)
                logger.debug(
                    (
                        f"Successfully calculated the freshness lifetime of the resource located at "
                        f"{get_safe_url(request.url)} using heuristics."
                    )
                )
            else:
                logger.debug(
                    (
                        "Could not calculate the freshness lifetime of "
                        f"the resource located at {get_safe_url(request.url)}. "
                        "Making a conditional request to revalidate the response."
                    )
                )
                # If Freshness cannot be calculated, then send the request
                self._make_request_conditional(request=request, response=response)
                return request

        age = get_age(response, self._clock)
        is_fresh = freshness_lifetime > age

        # The min-fresh request directive indicates that the client
        # prefers a response whose freshness lifetime is no less than
        #  its current age plus the specified time in seconds.
        # That is, the client wants a response that will still
        # be fresh for at least the specified number of seconds.
        if request_cache_control.min_fresh is not None:
            if freshness_lifetime < (age + request_cache_control.min_fresh):
                logger.debug(
                    (
                        f"Considering the resource located at {get_safe_url(request.url)} "
                        "as invalid for cache use since the time left for "
                        "freshness is less than the min-fresh directive."
                    )
                )
                return None

        # The max-stale request directive indicates that the
        # client will accept a response that has exceeded its freshness lifetime.
        # If a value is present, then the client is willing to accept a response
        # that has exceeded its freshness lifetime by no more than the specified
        # number of seconds. If no value is assigned to max-stale, then
        # the client will accept a stale response of any age.
        if not is_fresh and request_cache_control.max_stale is not None:
            exceeded_freshness_lifetime = age - freshness_lifetime

            if request_cache_control.max_stale < exceeded_freshness_lifetime:
                logger.debug(
                    (
                        f"Considering the resource located at {get_safe_url(request.url)} "
                        "as invalid for cache use since the freshness lifetime has been exceeded more than max-stale."
                    )
                )
                return None
            else:
                logger.debug(
                    (
                        f"Considering the resource located at {get_safe_url(request.url)} "
                        "as valid for cache use since the freshness lifetime has been exceeded less than max-stale."
                    )
                )
                return response

        # The max-age request directive indicates that
        # the client prefers a response whose age is
        # less than or equal to the specified number of seconds.
        # Unless the max-stale request directive is also present,
        # the client does not wish to receive a stale response.
        if request_cache_control.max_age is not None:
            if request_cache_control.max_age < age:
                logger.debug(
                    (
                        f"Considering the resource located at {get_safe_url(request.url)} "
                        "as invalid for cache use since the age of the response exceeds the max-age directive."
                    )
                )
                return None

        # the stored response is one of the following:
        #   fresh (see Section 4.2), or
        #   allowed to be served stale (see Section 4.2.4), or
        #   successfully validated (see Section 4.3).
        if is_fresh:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as valid for cache use since it is fresh."
                )
            )
            return response
        else:
            logger.debug(
                (
                    f"Considering the resource located at {get_safe_url(request.url)} "
                    "as needing revalidation since it is not fresh."
                )
            )
            # Otherwise, make a conditional request
            self._make_request_conditional(request=request, response=response)
            return request

    def handle_validation_response(self, old_response: Response, new_response: Response) -> Response:
        """
        Handles incoming validation response.

        This method takes care of what to do with the incoming
        validation response; if it is a 304 response, it updates
        the headers with the new response and returns it.

        This method mirrors the relevant section from RFC 9111,
        see (https://www.rfc-editor.org/rfc/rfc9111.html#name-handling-a-validation-respo).
        """
        if new_response.status == 304:
            headers = get_updated_headers(
                stored_response_headers=old_response.headers,
                new_response_headers=new_response.headers,
            )
            old_response.headers = headers
            return old_response
        else:
            return new_response



import calendar
import hashlib
import time
import typing as tp
from email.utils import parsedate_tz

import anyio
import httpcore
import httpx

HEADERS_ENCODING = "iso-8859-1"


class BaseClock:
    def now(self) -> int:
        raise NotImplementedError()


class Clock(BaseClock):
    def now(self) -> int:
        return int(time.time())


def normalized_url(url: tp.Union[httpcore.URL, str, bytes]) -> str:
    if isinstance(url, str):  # pragma: no cover
        return url

    if isinstance(url, bytes):  # pragma: no cover
        return url.decode("ascii")

    if isinstance(url, httpcore.URL):
        port = f":{url.port}" if url.port is not None else ""
        return f"{url.scheme.decode('ascii')}://{url.host.decode('ascii')}{port}{url.target.decode('ascii')}"
    assert False, "Invalid type for `normalized_url`"  # pragma: no cover


def get_safe_url(url: httpcore.URL) -> str:
    httpx_url = httpx.URL(bytes(url).decode("ascii"))

    schema = httpx_url.scheme
    host = httpx_url.host
    path = httpx_url.path

    return f"{schema}://{host}{path}"


def generate_key(request: httpcore.Request, body: bytes = b"") -> str:
    encoded_url = normalized_url(request.url).encode("ascii")

    key_parts = [request.method, encoded_url, body]

    # FIPs mode disables blake2 algorithm, use sha256 instead when not found.
    blake2b_hasher = None
    sha256_hasher = hashlib.sha256(usedforsecurity=False)
    try:
        blake2b_hasher = hashlib.blake2b(digest_size=16, usedforsecurity=False)
    except (ValueError, TypeError, AttributeError):
        pass

    hexdigest: str
    if blake2b_hasher:
        for part in key_parts:
            blake2b_hasher.update(part)

        hexdigest = blake2b_hasher.hexdigest()
    else:
        for part in key_parts:
            sha256_hasher.update(part)

        hexdigest = sha256_hasher.hexdigest()
    return hexdigest


def extract_header_values(
    headers: tp.List[tp.Tuple[bytes, bytes]],
    header_key: tp.Union[bytes, str],
    single: bool = False,
) -> tp.List[bytes]:
    if isinstance(header_key, str):
        header_key = header_key.encode(HEADERS_ENCODING)
    extracted_headers = []
    for key, value in headers:
        if key.lower() == header_key.lower():
            extracted_headers.append(value)
            if single:
                break
    return extracted_headers


def extract_header_values_decoded(
    headers: tp.List[tp.Tuple[bytes, bytes]], header_key: bytes, single: bool = False
) -> tp.List[str]:
    values = extract_header_values(headers=headers, header_key=header_key, single=single)
    return [value.decode(HEADERS_ENCODING) for value in values]


def header_presents(headers: tp.List[tp.Tuple[bytes, bytes]], header_key: bytes) -> bool:
    return bool(extract_header_values(headers, header_key, single=True))


def parse_date(date: str) -> tp.Optional[int]:
    expires = parsedate_tz(date)
    if expires is None:
        return None
    timestamp = calendar.timegm(expires[:6])
    return timestamp


async def asleep(seconds: tp.Union[int, float]) -> None:
    await anyio.sleep(seconds)


def sleep(seconds: tp.Union[int, float]) -> None:
    time.sleep(seconds)


def float_seconds_to_int_milliseconds(seconds: float) -> int:
    return int(seconds * 1000)



from __future__ import annotations

import types
import typing as tp

from httpcore._async.interfaces import AsyncRequestInterface
from httpcore._exceptions import ConnectError
from httpcore._models import Request, Response

from .._controller import Controller, allowed_stale
from .._headers import parse_cache_control
from .._serializers import JSONSerializer, Metadata
from .._utils import extract_header_values_decoded
from ._storages import AsyncBaseStorage, AsyncFileStorage

T = tp.TypeVar("T")

__all__ = ("AsyncCacheConnectionPool",)


async def fake_stream(content: bytes) -> tp.AsyncIterable[bytes]:
    yield content


def generate_504() -> Response:
    return Response(status=504)


class AsyncCacheConnectionPool(AsyncRequestInterface):
    """An HTTP Core Connection Pool that supports HTTP caching.

    :param pool: `Connection Pool` that our class wraps in order to add an HTTP Cache layer on top of
    :type pool: AsyncRequestInterface
    :param storage: Storage that handles how the responses should be saved., defaults to None
    :type storage: tp.Optional[AsyncBaseStorage], optional
    :param controller: Controller that manages the cache behavior at the specification level, defaults to None
    :type controller: tp.Optional[Controller], optional
    """

    def __init__(
        self,
        pool: AsyncRequestInterface,
        storage: tp.Optional[AsyncBaseStorage] = None,
        controller: tp.Optional[Controller] = None,
    ) -> None:
        self._pool = pool

        self._storage = storage if storage is not None else AsyncFileStorage(serializer=JSONSerializer())

        if not isinstance(self._storage, AsyncBaseStorage):  # pragma: no cover
            raise TypeError(f"Expected subclass of `AsyncBaseStorage` but got `{storage.__class__.__name__}`")

        self._controller = controller if controller is not None else Controller()

    async def handle_async_request(self, request: Request) -> Response:
        """
        Handles HTTP requests while also implementing HTTP caching.

        :param request: An HTTP request
        :type request: httpcore.Request
        :return: An HTTP response
        :rtype: httpcore.Response
        """

        if request.extensions.get("cache_disabled", False):
            request.headers.extend([(b"cache-control", b"no-cache"), (b"cache-control", b"max-age=0")])

        if request.method.upper() not in [b"GET", b"HEAD"]:
            # If the HTTP method is, for example, POST,
            # we must also use the request data to generate the hash.
            assert isinstance(request.stream, tp.AsyncIterable)
            body_for_key = b"".join([chunk async for chunk in request.stream])
            request.stream = fake_stream(body_for_key)
        else:
            body_for_key = b""

        key = self._controller._key_generator(request, body_for_key)
        stored_data = await self._storage.retrieve(key)

        request_cache_control = parse_cache_control(extract_header_values_decoded(request.headers, b"Cache-Control"))

        if request_cache_control.only_if_cached and not stored_data:
            return generate_504()

        if stored_data:
            # Try using the stored response if it was discovered.

            stored_response, stored_request, metadata = stored_data

            # Immediately read the stored response to avoid issues when trying to access the response body.
            stored_response.read()

            res = self._controller.construct_response_from_cache(
                request=request,
                response=stored_response,
                original_request=stored_request,
            )

            if isinstance(res, Response):
                # Simply use the response if the controller determines it is ready for use.
                return await self._create_hishel_response(
                    key=key,
                    response=stored_response,
                    request=request,
                    metadata=metadata,
                    cached=True,
                    revalidated=False,
                )

            if request_cache_control.only_if_cached:
                return generate_504()

            if isinstance(res, Request):
                # Controller has determined that the response needs to be re-validated.

                try:
                    revalidation_response = await self._pool.handle_async_request(res)
                except ConnectError:
                    # If there is a connection error, we can use the stale response if allowed.
                    if self._controller._allow_stale and allowed_stale(response=stored_response):
                        return await self._create_hishel_response(
                            key=key,
                            response=stored_response,
                            request=request,
                            metadata=metadata,
                            cached=True,
                            revalidated=False,
                        )
                    raise  # pragma: no cover
                # Merge headers with the stale response.
                final_response = self._controller.handle_validation_response(
                    old_response=stored_response, new_response=revalidation_response
                )

                await final_response.aread()

                # RFC 9111: 4.3.3. Handling a Validation Response
                # A 304 (Not Modified) response status code indicates that the stored response can be updated and
                # reused. A full response (i.e., one containing content) indicates that none of the stored responses
                # nominated in the conditional request are suitable. Instead, the cache MUST use the full response to
                # satisfy the request. The cache MAY store such a full response, subject to its constraints.
                if revalidation_response.status != 304 and self._controller.is_cachable(
                    request=request, response=final_response
                ):
                    await self._storage.store(key, response=final_response, request=request)

                return await self._create_hishel_response(
                    key=key,
                    response=final_response,
                    request=request,
                    cached=revalidation_response.status == 304,
                    revalidated=True,
                    metadata=metadata,
                )

        regular_response = await self._pool.handle_async_request(request)
        await regular_response.aread()

        if self._controller.is_cachable(request=request, response=regular_response):
            await self._storage.store(key, response=regular_response, request=request)

        return await self._create_hishel_response(
            key=key, response=regular_response, request=request, cached=False, revalidated=False
        )

    async def _create_hishel_response(
        self,
        key: str,
        response: Response,
        request: Request,
        cached: bool,
        revalidated: bool,
        metadata: Metadata | None = None,
    ) -> Response:
        if cached:
            assert metadata
            metadata["number_of_uses"] += 1
            await self._storage.update_metadata(key=key, request=request, response=response, metadata=metadata)
            response.extensions["from_cache"] = True  # type: ignore[index]
            response.extensions["cache_metadata"] = metadata  # type: ignore[index]
        else:
            response.extensions["from_cache"] = False  # type: ignore[index]
        response.extensions["revalidated"] = revalidated  # type: ignore[index]
        return response

    async def aclose(self) -> None:
        await self._storage.aclose()

        if hasattr(self._pool, "aclose"):  # pragma: no cover
            await self._pool.aclose()

    async def __aenter__(self: T) -> T:
        return self

    async def __aexit__(
        self,
        exc_type: tp.Optional[tp.Type[BaseException]] = None,
        exc_value: tp.Optional[BaseException] = None,
        traceback: tp.Optional[types.TracebackType] = None,
    ) -> None:
        await self.aclose()



import pytest

from hishel import LFUCache


def test_lfu_cache():
    cache: LFUCache[int, int] = LFUCache(2)

    cache.put(1, 2)
    a = cache.get(1)

    assert a == 2


def test_lfu_cache_delete():
    cache: LFUCache[int, int] = LFUCache(2)

    cache.put(1, 2)
    cache.put(3, 4)
    cache.put(5, 6)
    cache.get(3)
    cache.get(3)

    with pytest.raises(KeyError):
        cache.get(1)


def test_lfu_cache_invalid_capacity():
    with pytest.raises(ValueError, match="Capacity must be positive"):
        LFUCache(0)


def test_lfu_cache_delete_least_frequently():
    cache: LFUCache[int, int] = LFUCache(2)

    cache.put(1, 10)
    cache.put(2, 10)

    cache.get(1)
    cache.put(3, 10)

    with pytest.raises(KeyError):
        cache.get(2)

    cache: LFUCache[int, int] = LFUCache(2)  # type: ignore[no-redef]

    cache.put(1, 10)
    cache.put(2, 10)

    cache.get(2)
    cache.put(3, 10)

    with pytest.raises(KeyError):
        cache.get(1)


def test_lfu_cache_remove_key():
    cache: LFUCache[int, int] = LFUCache(2)

    cache.put(1, 10)
    cache.put(2, 10)

    cache.remove_key(1)

    with pytest.raises(KeyError):
        cache.get(1)

    cache.put(1, 10)
    cache.get(1)

    assert cache.min_freq == 1
    cache.remove_key(2)
    assert cache.min_freq == 2


def test_lfu_cache_put_existing():
    cache: LFUCache[int, int] = LFUCache(2)

    cache.put(1, 10)
    cache.put(2, 10)

    cache.put(1, 20)

    assert cache.get(1) == 20

    cache: LFUCache[int, int] = LFUCache(2)  # type: ignore[no-redef]

    cache.put(1, 10)
    cache.put(1, 20)

    assert cache.get(1) == 20



import datetime
import os
from pathlib import Path

import anysqlite
import pytest
from httpcore import Request, Response

from hishel import AsyncFileStorage, AsyncInMemoryStorage, AsyncRedisStorage, AsyncSQLiteStorage
from hishel._serializers import Metadata
from hishel._utils import asleep, generate_key

dummy_metadata = Metadata(cache_key="test", number_of_uses=0, created_at=datetime.datetime.now(datetime.timezone.utc))


async def is_redis_down() -> bool:
    import redis.asyncio as redis

    connection = redis.Redis()
    try:
        return not await connection.ping()
    except BaseException:  # pragma: no cover
        return True


@pytest.mark.anyio
async def test_filestorage(use_temp_dir):
    storage = AsyncFileStorage()

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(key, response=response, request=request, metadata=dummy_metadata)

    stored_data = await storage.retrieve(key)
    assert stored_data is not None
    stored_response, stored_request, metadata = stored_data
    stored_response.read()
    assert isinstance(stored_response, Response)
    assert stored_response.status == 200
    assert stored_response.headers == []
    assert stored_response.content == b"test"


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_redisstorage(anyio_backend):
    if await is_redis_down():  # pragma: no cover
        pytest.fail("Redis server was not found")
    storage = AsyncRedisStorage()

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(key, response=response, request=request, metadata=dummy_metadata)

    stored_data = await storage.retrieve(key)
    assert stored_data is not None
    stored_response, stored_request, metadata = stored_data
    stored_response.read()
    assert isinstance(stored_response, Response)
    assert stored_response.status == 200
    assert stored_response.headers == []
    assert stored_response.content == b"test"


@pytest.mark.anyio
async def test_sqlitestorage():
    storage = AsyncSQLiteStorage(connection=await anysqlite.connect(":memory:"))

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(key, response=response, request=request, metadata=dummy_metadata)

    stored_data = await storage.retrieve(key)
    assert stored_data is not None
    stored_response, stored_request, metadata = stored_data
    stored_response.read()
    assert isinstance(stored_response, Response)
    assert stored_response.status == 200
    assert stored_response.headers == []
    assert stored_response.content == b"test"


@pytest.mark.anyio
async def test_inmemorystorage():
    storage = AsyncInMemoryStorage()

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(key, response=response, request=request, metadata=dummy_metadata)

    stored_data = await storage.retrieve(key)
    assert stored_data is not None
    stored_response, stored_request, metadata = stored_data
    stored_response.read()
    assert isinstance(stored_response, Response)
    assert stored_response.status == 200
    assert stored_response.headers == []
    assert stored_response.content == b"test"


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_filestorage_expired(use_temp_dir, anyio_backend):
    storage = AsyncFileStorage(ttl=0.2, check_ttl_every=0.1)
    first_request = Request(b"GET", "https://example.com")
    second_request = Request(b"GET", "https://anotherexample.com")

    first_key = generate_key(first_request)
    second_key = generate_key(second_request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(first_key, response=response, request=first_request, metadata=dummy_metadata)
    assert await storage.retrieve(first_key) is not None

    await asleep(0.3)
    await storage.store(second_key, response=response, request=second_request, metadata=dummy_metadata)

    assert await storage.retrieve(first_key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_filestorage_timer(use_temp_dir, anyio_backend):
    storage = AsyncFileStorage(ttl=0.2, check_ttl_every=0.2)

    first_request = Request(b"GET", "https://example.com")
    second_request = Request(b"GET", "https://anotherexample.com")

    first_key = generate_key(first_request)
    second_key = generate_key(second_request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(first_key, response=response, request=first_request, metadata=dummy_metadata)
    assert await storage.retrieve(first_key) is not None
    await asleep(0.1)
    assert await storage.retrieve(first_key) is not None
    await storage.store(second_key, response=response, request=second_request, metadata=dummy_metadata)
    assert await storage.retrieve(second_key) is not None
    await asleep(0.1)
    assert await storage.retrieve(first_key) is None
    assert await storage.retrieve(second_key) is not None
    await asleep(0.1)
    assert await storage.retrieve(second_key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_filestorage_ttl_after_hits(use_temp_dir, anyio_backend):
    storage = AsyncFileStorage(ttl=0.2, check_ttl_every=0.2)

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    # Storing
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.08 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.16 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.24 second
    await asleep(0.08)
    assert await storage.retrieve(key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_redisstorage_expired(anyio_backend):
    if await is_redis_down():  # pragma: no cover
        pytest.fail("Redis server was not found")
    storage = AsyncRedisStorage(ttl=0.1)
    first_request = Request(b"GET", "https://example.com")
    second_request = Request(b"GET", "https://anotherexample.com")

    first_key = generate_key(first_request)
    second_key = generate_key(second_request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(first_key, response=response, request=first_request, metadata=dummy_metadata)
    assert await storage.retrieve(first_key) is not None

    await asleep(0.3)
    await storage.store(second_key, response=response, request=second_request, metadata=dummy_metadata)

    assert await storage.retrieve(first_key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_redis_ttl_after_hits(use_temp_dir, anyio_backend):
    storage = AsyncRedisStorage(ttl=0.2)

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    # Storing
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.08 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.16 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.24 second
    await asleep(0.08)
    assert await storage.retrieve(key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_sqlite_expired(anyio_backend):
    storage = AsyncSQLiteStorage(ttl=0.1, connection=await anysqlite.connect(":memory:"))
    first_request = Request(b"GET", "https://example.com")
    second_request = Request(b"GET", "https://anotherexample.com")

    first_key = generate_key(first_request)
    second_key = generate_key(second_request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(first_key, response=response, request=first_request, metadata=dummy_metadata)
    assert await storage.retrieve(first_key) is not None

    await asleep(0.3)
    await storage.store(second_key, response=response, request=second_request, metadata=dummy_metadata)

    assert await storage.retrieve(first_key) is None


@pytest.mark.xfail
@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_sqlite_ttl_after_hits(use_temp_dir, anyio_backend):
    storage = AsyncSQLiteStorage(ttl=0.2)

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    # Storing
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.08 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.16 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.24 second
    await asleep(0.08)
    assert await storage.retrieve(key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_inmemory_expired(anyio_backend):
    storage = AsyncInMemoryStorage(ttl=0.1)
    first_request = Request(b"GET", "https://example.com")
    second_request = Request(b"GET", "https://anotherexample.com")

    first_key = generate_key(first_request)
    second_key = generate_key(second_request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    await storage.store(first_key, response=response, request=first_request, metadata=dummy_metadata)
    assert await storage.retrieve(first_key) is not None

    await asleep(0.3)
    await storage.store(second_key, response=response, request=second_request, metadata=dummy_metadata)

    assert await storage.retrieve(first_key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_inmemory_ttl_after_hits(use_temp_dir, anyio_backend):
    storage = AsyncInMemoryStorage(ttl=0.2)

    request = Request(b"GET", "https://example.com")

    key = generate_key(request)

    response = Response(200, headers=[], content=b"test")
    await response.aread()

    # Storing
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.08 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.16 second
    await asleep(0.08)
    await storage.update_metadata(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None

    # Retrieving after 0.24 second
    await asleep(0.08)
    assert await storage.retrieve(key) is None


@pytest.mark.anyio
async def test_filestorage_empty_file_exception(use_temp_dir):
    """When working with concurrency sometimes Hishel
    may leave a cache file empty. In this case this should not
    cause a `JSONDecodeError`, but treat this situation as
    no cache file was created. Issue #180"""

    storage = AsyncFileStorage()
    request = Request(b"GET", "https://example.com")
    key = generate_key(request)
    response = Response(200, headers=[], content=b"test")
    await response.aread()
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    stored_data = await storage.retrieve(key)

    assert stored_data is not None
    filedir = Path(os.getcwd() + "/.cache/hishel/" + key)
    with open(filedir, "w+", encoding="utf-8") as file:
        file.truncate(0)
    assert os.path.getsize(filedir) == 0
    assert await storage.retrieve(key) is None


@pytest.mark.anyio
async def test_filestorage_remove(use_temp_dir):
    storage = AsyncFileStorage()
    request = Request(b"GET", "https://example.com")

    key = generate_key(request)
    response = Response(200, headers=[], content=b"test")

    await response.aread()
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None
    await storage.remove(key)
    assert await storage.retrieve(key) is None


@pytest.mark.parametrize("anyio_backend", ["asyncio"])
async def test_redisstorage_remove(anyio_backend):
    if await is_redis_down():  # pragma: no cover
        pytest.fail("Redis server was not found")

    storage = AsyncRedisStorage()
    request = Request(b"GET", "https://example.com")

    key = generate_key(request)
    response = Response(200, headers=[], content=b"test")

    await response.aread()
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None
    await storage.remove(key)
    assert await storage.retrieve(key) is None


@pytest.mark.anyio
async def test_sqlitestorage_remove():
    storage = AsyncSQLiteStorage(connection=await anysqlite.connect(":memory:"))
    request = Request(b"GET", "https://example.com")

    key = generate_key(request)
    response = Response(200, headers=[], content=b"test")

    await response.aread()
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None
    await storage.remove(key)
    assert await storage.retrieve(key) is None


@pytest.mark.anyio
async def test_inmemorystorage_remove():
    storage = AsyncInMemoryStorage()
    request = Request(b"GET", "https://example.com")

    key = generate_key(request)
    response = Response(200, headers=[], content=b"test")

    await response.aread()
    await storage.store(key, response=response, request=request, metadata=dummy_metadata)
    assert await storage.retrieve(key) is not None
    await storage.remove(key)
    assert await storage.retrieve(key) is None



#!venv/bin/python
import os
import re
import sys

SUBS = [
    ("async def", "def"),
    ("async with", "with"),
    ("await ", ""),
    ("async for", "for"),
    ("__aiter__", "__iter__"),
    ("AsyncIterator", "Iterator"),
    ("AsyncFileStorage", "FileStorage"),
    ("AsyncBaseStorage", "BaseStorage"),
    ("AsyncFileManager", "FileManager"),
    ("MockAsyncConnectionPool", "MockConnectionPool"),
    ("MockAsyncTransport", "MockTransport"),
    ("AsyncRedisStorage", "RedisStorage"),
    ("AsyncSQLiteStorage", "SQLiteStorage"),
    ("AsyncInMemoryStorage", "InMemoryStorage"),
    ("AsyncS3Storage", "S3Storage"),
    ("AsyncS3Manager", "S3Manager"),
    ("import redis.asyncio as redis", "import redis"),
    ("AsyncCacheTransport", "CacheTransport"),
    ("AsyncBaseTransport", "BaseTransport"),
    ("AsyncCacheClient", "CacheClient"),
    ("AsyncClient", "Client"),
    ("AsyncIterable", "Iterable"),
    ("AsyncCacheStream", "CacheStream"),
    ("AsyncByteStream", "SyncByteStream"),
    ("AsyncCacheConnectionPool", "CacheConnectionPool"),
    ("handle_async_request", "handle_request"),
    ("aread", "read"),
    ("aclose", "close"),
    ("asleep", "sleep"),
    ("AsyncLock", "Lock"),
    (
        "from httpcore._async.interfaces import AsyncRequestInterface",
        "from httpcore._sync.interfaces import RequestInterface",
    ),
    ("from hishel._async._transports", "from hishel._sync._transports"),
    ("AsyncRequestInterface", "RequestInterface"),
    ("__aenter__", "__enter__"),
    ("__aexit__", "__exit__"),
    ("*@pytest.mark.anyio", ""),
    (r'*@pytest.mark.parametrize\("anyio_backend", \["asyncio"\]\)', ""),
    ("anysqlite", "sqlite3"),
]
COMPILED_SUBS = [(re.compile(r"(^|\b)" + regex + r"($|\b)"), repl) for regex, repl in SUBS]

USED_SUBS = set()


def unasync_line(line):
    for index, (regex, repl) in enumerate(COMPILED_SUBS):
        old_line = line
        line = re.sub(regex, repl, line)
        if index not in USED_SUBS:
            if line != old_line:
                USED_SUBS.add(index)
    return line


def unasync_file(in_path, out_path):
    with open(in_path) as in_file:
        with open(out_path, "w", newline="") as out_file:
            for line in in_file.readlines():
                line = unasync_line(line)
                out_file.write(line)


def unasync_file_check(in_path, out_path):
    with open(in_path) as in_file:
        with open(out_path) as out_file:
            for in_line, out_line in zip(in_file.readlines(), out_file.readlines()):
                expected = unasync_line(in_line)
                if out_line != expected:
                    print(f"unasync mismatch between {in_path!r} and {out_path!r}")
                    print(f"Async code:         {in_line!r}")
                    print(f"Expected sync code: {expected!r}")
                    print(f"Actual sync code:   {out_line!r}")
                    sys.exit(1)


def unasync_dir(in_dir, out_dir, check_only=False):
    for dirpath, dirnames, filenames in os.walk(in_dir):
        for filename in filenames:
            if not filename.endswith(".py"):
                continue
            rel_dir = os.path.relpath(dirpath, in_dir)
            in_path = os.path.normpath(os.path.join(in_dir, rel_dir, filename))
            out_path = os.path.normpath(os.path.join(out_dir, rel_dir, filename))
            print(in_path, "->", out_path)
            if check_only:
                unasync_file_check(in_path, out_path)
            else:
                unasync_file(in_path, out_path)


def main():
    check_only = "--check" in sys.argv
    unasync_dir("hishel/_async", "hishel/_sync", check_only=check_only)
    unasync_dir("tests/_async", "tests/_sync", check_only=check_only)

    if len(USED_SUBS) != len(SUBS):
        unused_subs = [SUBS[i] for i in range(len(SUBS)) if i not in USED_SUBS]

        from pprint import pprint

        print("This SUBS was not used")
        pprint(unused_subs)
        exit(1)


if __name__ == "__main__":
    main()



from unittest import mock

import httpcore
import pytest
from httpcore import Request

from hishel._controller import get_updated_headers
from hishel._utils import (
    extract_header_values,
    extract_header_values_decoded,
    float_seconds_to_int_milliseconds,
    generate_key,
    get_safe_url,
    header_presents,
    parse_date,
)


def test_generate_key():
    request = Request(b"GET", "https://example.com", headers=[])

    key = generate_key(request)

    assert key == "bd152069787aaad359c85af6f2edbb25"


def test_fips_generate_key():
    request = Request(b"GET", "https://example.com", headers=[])

    # Simulate FIPS mode by using sha256 instead of blake2b
    with mock.patch("hashlib.blake2b", side_effect=AttributeError("ERROR")):
        key = generate_key(request)

    assert key == "ea96dc6995764a0e6cf26bd2550deb01c18f69c0e586aa1fe201683129b8c15a"


def test_extract_header_values():
    headers = [
        (b"Content-Type", b"application/json"),
        (b"Content-Type", b"application/html"),
    ]

    values = extract_header_values(headers, b"Content-Type")
    assert values == [b"application/json", b"application/html"]


def test_extract_header_values_decoded():
    headers = [
        (b"Content-Type", b"application/json"),
        (b"Content-Type", b"application/html"),
    ]

    values = extract_header_values_decoded(headers, b"Content-Type")
    assert values == ["application/json", "application/html"]


def test_extract_header_single_value():
    headers = [
        (b"Content-Type", b"application/json"),
        (b"Content-Type", b"application/html"),
    ]

    values = extract_header_values(headers, b"Content-Type", single=True)
    assert values == [b"application/json"]


def test_header_presents():
    headers = [
        (b"Content-Type", b"application/json"),
        (b"Content-Type", b"application/html"),
        (b"Accept", b"application/json"),
    ]

    accept_presents = header_presents(headers, b"Accept")
    assert accept_presents

    transfer_encoding_presents = header_presents(headers, b"Transfer-Encoding")
    assert not transfer_encoding_presents


def test_get_updated_headers():
    old_headers = [(b"Content-Type", b"application/json"), (b"Language", b"en")]

    new_headers = [
        (b"Language", b"am"),
        (b"Content-Length", b"1024"),
        (b"Authorization", b"secret-key"),
    ]

    update_headers = get_updated_headers(stored_response_headers=old_headers, new_response_headers=new_headers)

    assert len(update_headers) == 3
    assert extract_header_values(update_headers, b"Language")[0] == b"am"
    assert extract_header_values(update_headers, b"Content-Type")[0] == b"application/json"
    assert extract_header_values(update_headers, b"Authorization")[0] == b"secret-key"


def test_parse_date():
    date = "Mon, 25 Aug 2015 12:00:00 GMT"
    timestamp = parse_date(date)
    assert timestamp == 1440504000


def test_parse_invalid_date():
    date = "0"
    timestamp = parse_date(date)
    assert timestamp is None


def test_float_seconds_to_milliseconds():
    seconds = 1.234
    milliseconds = float_seconds_to_int_milliseconds(seconds)
    assert milliseconds == 1234


@pytest.mark.parametrize(
    "url, expected",
    [
        pytest.param(
            "https://example.com/path?query=1",
            "https://example.com/path",
            id="url_with_query_is_ignored",
        ),
        pytest.param(
            "https://example.com/path",
            "https://example.com/path",
            id="url_without_query",
        ),
        pytest.param("https://example.com", "https://example.com/", id="url_without_path"),
        pytest.param(
            "https://xn--e1afmkfd.xn--p1ag",
            "https://пример.ру/",
            id="url_with_idna",
        ),
    ],
)
def test_safe_url(
    url: str,
    expected: str,
) -> None:
    httpcore_url = httpcore.URL(url)
    safe_url = get_safe_url(httpcore_url)

    assert safe_url == expected



import datetime

from httpcore import Request, Response

from hishel._serializers import (
    JSONSerializer,
    Metadata,
    PickleSerializer,
    YAMLSerializer,
)
from hishel._utils import normalized_url


def test_pickle_serializer_dumps_and_loads():
    request = Request(
        method="GET",
        url="https://example.com",
        headers=[(b"Accept-Encoding", b"gzip")],
        extensions={"sni_hostname": "example.com"},
    )
    response = Response(
        status=200,
        headers=[
            (b"Content-Type", b"application/json"),
            (b"Transfer-Encoding", b"chunked"),
        ],
        content=b"test",
        extensions={"reason_phrase": b"OK", "http_version": b"HTTP/1.1"},
    )
    response.read()

    metadata = Metadata(
        cache_key="test",
        number_of_uses=0,
        created_at=datetime.datetime(year=2003, month=8, day=25, hour=12),
    )

    raw_response = PickleSerializer().dumps(response=response, request=request, metadata=metadata)

    response, request, metadata = PickleSerializer().loads(raw_response)
    response.read()
    assert response.status == 200
    assert response.headers == [
        (b"Content-Type", b"application/json"),
        (b"Transfer-Encoding", b"chunked"),
    ]
    assert response.content == b"test"
    assert response.extensions == {"http_version": b"HTTP/1.1", "reason_phrase": b"OK"}

    assert request.method == b"GET"
    assert normalized_url(request.url) == "https://example.com/"
    assert request.headers == [(b"Accept-Encoding", b"gzip")]
    assert request.extensions == {"sni_hostname": "example.com"}

    assert metadata["cache_key"] == "test"
    assert metadata["number_of_uses"] == 0
    assert metadata["created_at"] == datetime.datetime(year=2003, month=8, day=25, hour=12)


def test_dict_serializer_dumps():
    request = Request(
        method="GET",
        url="https://example.com",
        headers=[(b"Accept-Encoding", b"gzip")],
        extensions={"sni_hostname": "example.com"},
    )
    response = Response(
        status=200,
        headers=[
            (b"Content-Type", b"application/json"),
            (b"Transfer-Encoding", b"chunked"),
        ],
        content=b"test",
        extensions={"reason_phrase": b"OK", "http_version": b"HTTP/1.1"},
    )
    response.read()

    metadata = Metadata(
        cache_key="test",
        number_of_uses=0,
        created_at=datetime.datetime(year=2003, month=8, day=25, hour=12),
    )

    full_json = JSONSerializer().dumps(response=response, request=request, metadata=metadata)

    assert full_json == "\n".join(
        [
            "{",
            '    "response": {',
            '        "status": 200,',
            '        "headers": [',
            "            [",
            '                "Content-Type",',
            '                "application/json"',
            "            ],",
            "            [",
            '                "Transfer-Encoding",',
            '                "chunked"',
            "            ]",
            "        ],",
            '        "content": "dGVzdA==",',
            '        "extensions": {',
            '            "reason_phrase": "OK",',
            '            "http_version": "HTTP/1.1"',
            "        }",
            "    },",
            '    "request": {',
            '        "method": "GET",',
            '        "url": "https://example.com/",',
            '        "headers": [',
            "            [",
            '                "Accept-Encoding",',
            '                "gzip"',
            "            ]",
            "        ],",
            '        "extensions": {',
            '            "sni_hostname": "example.com"',
            "        }",
            "    },",
            '    "metadata": {',
            '        "cache_key": "test",',
            '        "number_of_uses": 0,',
            '        "created_at": "Mon, 25 Aug 2003 12:00:00 GMT"',
            "    }",
            "}",
        ]
    )


def test_dict_serializer_loads():
    raw_response = "\n".join(
        [
            "{",
            '    "response": {',
            '        "status": 200,',
            '        "headers": [',
            "            [",
            '                "Content-Type",',
            '                "application/json"',
            "            ],",
            "            [",
            '                "Transfer-Encoding",',
            '                "chunked"',
            "            ]",
            "        ],",
            '        "content": "dGVzdA==",',
            '        "extensions": {',
            '            "reason_phrase": "OK",',
            '            "http_version": "HTTP/1.1"',
            "        }",
            "    },",
            '    "request": {',
            '        "method": "GET",',
            '        "url": "https://example.com/",',
            '        "headers": [',
            "            [",
            '                "Accept-Encoding",',
            '                "gzip"',
            "            ]",
            "        ],",
            '        "extensions": {',
            '            "sni_hostname": "example.com"',
            "        }",
            "    },",
            '    "metadata": {',
            '        "cache_key": "test",',
            '        "number_of_uses": 0,',
            '        "created_at": "Mon, 25 Aug 2003 12:00:00 GMT"',
            "    }",
            "}",
        ]
    )

    response, request, metadata = JSONSerializer().loads(raw_response)
    response.read()
    assert response.status == 200
    assert response.headers == [
        (b"Content-Type", b"application/json"),
        (b"Transfer-Encoding", b"chunked"),
    ]
    assert response.content == b"test"
    assert response.extensions == {"http_version": b"HTTP/1.1", "reason_phrase": b"OK"}

    assert request.method == b"GET"
    assert normalized_url(request.url) == "https://example.com/"
    assert request.headers == [(b"Accept-Encoding", b"gzip")]
    assert request.extensions == {"sni_hostname": "example.com"}

    assert metadata["cache_key"] == "test"
    assert metadata["number_of_uses"] == 0
    assert metadata["created_at"] == datetime.datetime(year=2003, month=8, day=25, hour=12)


def test_yaml_serializer_dumps():
    request = Request(
        method="GET",
        url="https://example.com",
        headers=[(b"Accept-Encoding", b"gzip")],
        extensions={"sni_hostname": "example.com"},
    )
    response = Response(
        status=200,
        headers=[
            (b"Content-Type", b"application/json"),
            (b"Transfer-Encoding", b"chunked"),
        ],
        content=b"test",
        extensions={"reason_phrase": b"OK", "http_version": b"HTTP/1.1"},
    )
    response.read()

    metadata = Metadata(
        cache_key="test",
        number_of_uses=0,
        created_at=datetime.datetime(year=2003, month=8, day=25, hour=12),
    )

    full_json = YAMLSerializer().dumps(response=response, request=request, metadata=metadata)

    assert full_json == "\n".join(
        [
            "response:",
            "  status: 200",
            "  headers:",
            "  - - Content-Type",
            "    - application/json",
            "  - - Transfer-Encoding",
            "    - chunked",
            "  content: dGVzdA==",
            "  extensions:",
            "    reason_phrase: OK",
            "    http_version: HTTP/1.1",
            "request:",
            "  method: GET",
            "  url: https://example.com/",
            "  headers:",
            "  - - Accept-Encoding",
            "    - gzip",
            "  extensions:",
            "    sni_hostname: example.com",
            "metadata:",
            "  cache_key: test",
            "  number_of_uses: 0",
            "  created_at: Mon, 25 Aug 2003 12:00:00 GMT",
            "",
        ]
    )


def test_yaml_serializer_loads():
    raw_response = "\n".join(
        [
            "response:",
            "  status: 200",
            "  headers:",
            "  - - Content-Type",
            "    - application/json",
            "  - - Transfer-Encoding",
            "    - chunked",
            "  content: dGVzdA==",
            "  extensions:",
            "    reason_phrase: OK",
            "    http_version: HTTP/1.1",
            "request:",
            "  method: GET",
            "  url: https://example.com/",
            "  headers:",
            "  - - Accept-Encoding",
            "    - gzip",
            "  extensions:",
            "    sni_hostname: example.com",
            "metadata:",
            "  cache_key: test",
            "  number_of_uses: 0",
            "  created_at: Mon, 25 Aug 2003 12:00:00 GMT",
            "",
        ]
    )

    response, request, metadata = YAMLSerializer().loads(raw_response)
    response.read()
    assert response.status == 200
    assert response.headers == [
        (b"Content-Type", b"application/json"),
        (b"Transfer-Encoding", b"chunked"),
    ]
    assert response.content == b"test"
    assert response.extensions == {"http_version": b"HTTP/1.1", "reason_phrase": b"OK"}

    assert request.method == b"GET"
    assert normalized_url(request.url) == "https://example.com/"
    assert request.headers == [(b"Accept-Encoding", b"gzip")]
    assert request.extensions == {"sni_hostname": "example.com"}

    assert metadata["cache_key"] == "test"
    assert metadata["number_of_uses"] == 0
    assert metadata["created_at"] == datetime.datetime(year=2003, month=8, day=25, hour=12)



import httpx
import pytest

import hishel
from hishel._utils import BaseClock, extract_header_values_decoded



def test_transport_301():
    with hishel.MockTransport() as transport:
        transport.add_responses([httpx.Response(301, headers=[(b"Location", b"https://example.com")])])
        with hishel.CacheTransport(
            transport=transport, storage=hishel.InMemoryStorage()
        ) as cache_transport:
            request = httpx.Request("GET", "https://www.example.com")

            cache_transport.handle_request(request)
            response = cache_transport.handle_request(request)
            assert response.extensions["from_cache"]



def test_transport_response_validation():
    with hishel.MockTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=3600"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                    content="test",
                ),
                httpx.Response(
                    304,
                    headers=[
                        (b"Cache-Control", b"max-age=3600"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                        (b"Content-Type", b"application/json"),
                    ],
                ),
            ]
        )
    with hishel.CacheTransport(
        transport=transport, storage=hishel.InMemoryStorage()
    ) as cache_transport:
        request = httpx.Request("GET", "https://www.example.com")

        cache_transport.handle_request(request)
        response = cache_transport.handle_request(request)
        assert response.status_code == 200
        assert response.extensions["from_cache"]
        assert response.extensions["revalidated"]
        assert "Content-Type" in response.headers
        assert response.headers["Content-Type"] == "application/json"
        assert response.read() == b"test"



def test_transport_stale_response():
    controller = hishel.Controller(allow_stale=True)

    with hishel.MockTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=3600"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                ),
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=3600"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                ),
            ]
        )
        with hishel.CacheTransport(
            transport=transport, controller=controller, storage=hishel.InMemoryStorage()
        ) as cache_transport:
            request = httpx.Request("GET", "https://www.example.com")
            cache_transport.handle_request(request)
            response = cache_transport.handle_request(request)
            assert not response.extensions["from_cache"]



def test_transport_stale_response_with_connecterror():
    controller = hishel.Controller(allow_stale=True)

    class ConnectErrorTransport(hishel.MockTransport):
        def handle_request(self, request: httpx.Request) -> httpx.Response:
            if not hasattr(self, "not_first_request"):
                setattr(self, "not_first_request", object())
                return super().handle_request(request)
            raise httpx._exceptions.ConnectError("test")

    with ConnectErrorTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=3600"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                ),
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=3600"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                ),
            ]
        )
        with hishel.CacheTransport(
            transport=transport, controller=controller, storage=hishel.InMemoryStorage()
        ) as cache_transport:
            request = httpx.Request("GET", "https://www.example.com")
            cache_transport.handle_request(request)
            response = cache_transport.handle_request(request)
            assert response.extensions["from_cache"]



def test_transport_with_only_if_cached_directive_without_stored_response():
    controller = hishel.Controller()

    with hishel.MockTransport() as transport:
        with hishel.CacheTransport(
            transport=transport, controller=controller, storage=hishel.InMemoryStorage()
        ) as cache_transport:
            response = cache_transport.handle_request(
                httpx.Request(
                    "GET",
                    "https://www.example.com",
                    headers=[(b"Cache-Control", b"only-if-cached")],
                )
            )
            assert response.status_code == 504



def test_transport_with_only_if_cached_directive_with_stored_response():
    controller = hishel.Controller()

    with hishel.MockTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=3600"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                    content=b"test",
                ),
            ]
        )
        with hishel.CacheTransport(
            transport=transport, controller=controller, storage=hishel.InMemoryStorage()
        ) as cache_transport:
            cache_transport.handle_request(httpx.Request("GET", "https://www.example.com"))
            response = cache_transport.handle_request(
                httpx.Request(
                    "GET",
                    "https://www.example.com",
                    headers=[(b"Cache-Control", b"only-if-cached")],
                )
            )
            assert response.status_code == 504



def test_transport_with_cache_disabled_extension():
    class MockedClock(BaseClock):
        def now(self) -> int:
            return 1440504001  # Mon, 25 Aug 2015 12:00:01 GMT

    cachable_response = httpx.Response(
        200,
        headers=[
            (b"Cache-Control", b"max-age=3600"),
            (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),  # 1 second before the clock
        ],
    )

    with hishel.MockTransport() as transport:
        transport.add_responses([cachable_response, httpx.Response(201)])
        with hishel.CacheTransport(
            transport=transport,
            controller=hishel.Controller(clock=MockedClock()),
            storage=hishel.InMemoryStorage(),
        ) as cache_transport:
            request = httpx.Request("GET", "https://www.example.com")
            # This should create a cache entry
            cache_transport.handle_request(request)
            # This should return from cache
            response = cache_transport.handle_request(request)
            assert response.extensions["from_cache"]
            # This should ignore the cache
            caching_disabled_request = httpx.Request(
                "GET", "https://www.example.com", extensions={"cache_disabled": True}
            )
            response = cache_transport.handle_request(caching_disabled_request)
            assert not response.extensions["from_cache"]
            assert response.status_code == 201



def test_transport_with_custom_key_generator():
    controller = hishel.Controller(key_generator=lambda request, body: request.url.host.decode())

    with hishel.MockTransport() as transport:
        transport.add_responses([httpx.Response(301)])
        with hishel.CacheTransport(
            transport=transport,
            controller=controller,
            storage=hishel.InMemoryStorage(),
        ) as cache_transport:
            request = httpx.Request("GET", "https://www.example.com")
            # This should create a cache entry
            cache_transport.handle_request(request)
            # This should return from cache
            response = cache_transport.handle_request(request)
            assert response.extensions["from_cache"]
            assert response.extensions["cache_metadata"]["cache_key"] == "www.example.com"



def test_transport_caching_post_method():
    controller = hishel.Controller(cacheable_methods=["POST"])

    with hishel.MockTransport() as transport:
        transport.add_responses([httpx.Response(301), httpx.Response(200)])
        with hishel.CacheTransport(
            transport=transport,
            controller=controller,
            storage=hishel.InMemoryStorage(),
        ) as cache_transport:
            request = httpx.Request("POST", "https://www.example.com", json={"request": 1})
            # This should create a cache entry
            cache_transport.handle_request(request)
            # This should return from cache
            response = cache_transport.handle_request(request)
            assert response.extensions["from_cache"]

            # Method and URL are the same but the body is different
            request = httpx.Request("POST", "https://www.example.com", json={"request": 2})

            # This should create a new cache entry instead of using the previous one
            response = cache_transport.handle_request(request)
            assert response.status_code == 200
            assert not response.extensions["from_cache"]



def test_revalidation_with_new_content():
    class MockedClock(BaseClock):
        current = 1440504000  # Mon, 25 Aug 2015 12:00:00 GMT

        def now(self) -> int:
            return self.current

    clock = MockedClock()
    controller = hishel.Controller(clock=clock)
    storage = hishel.InMemoryStorage()

    with hishel.MockTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=1"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                    content=b"Hello, World.",
                ),
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=10"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:01 GMT"),
                    ],
                    content=b"Eat at Joe's.",
                ),
                httpx.Response(
                    304,
                    headers=[
                        (b"Cache-Control", b"max-age=10"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:11 GMT"),
                    ],
                ),
            ]
        )
        with hishel.CacheTransport(
            transport=transport, controller=controller, storage=storage
        ) as cache_transport:
            # Miss, 200, store
            response = cache_transport.handle_request(httpx.Request("GET", "https://example.com/"))
            assert not response.extensions["from_cache"]

            # Hit
            response = cache_transport.handle_request(httpx.Request("GET", "https://example.com/"))
            assert response.extensions["from_cache"]
            assert response.extensions["cache_metadata"]["number_of_uses"] == 1

            # Cache contains the first response content
            stored = storage.retrieve(response.extensions["cache_metadata"]["cache_key"])
            assert stored
            stored_response, stored_request, stored_metadata = stored
            assert extract_header_values_decoded(stored_response.headers, b"Date") == ["Mon, 25 Aug 2015 12:00:00 GMT"]
            assert stored_response.content == b"Hello, World."

            # tic, tac... one second passed
            clock.current += 1

            # Miss (expired), send revalidation, 200, store
            response = cache_transport.handle_request(httpx.Request("GET", "https://example.com/"))
            assert not response.extensions["from_cache"]

            # Hit (cf issue #239)
            response = cache_transport.handle_request(httpx.Request("GET", "https://example.com/"))
            assert response.extensions["from_cache"]
            assert response.extensions["cache_metadata"]["number_of_uses"] == 1

            # Cache was updated and contains the second response content
            stored = storage.retrieve(response.extensions["cache_metadata"]["cache_key"])
            assert stored
            stored_response, stored_request, stored_metadata = stored
            assert extract_header_values_decoded(stored_response.headers, b"Date") == ["Mon, 25 Aug 2015 12:00:01 GMT"]
            assert stored_response.content == b"Eat at Joe's."

            # tic, tac, tic, tac... ten more seconds passed, let's check the 304 behavious is not broken
            clock.current += 10

            # Miss (expired), send revalidation, 304, update metadata but keep previous content
            response = cache_transport.handle_request(httpx.Request("GET", "https://example.com/"))
            assert response.extensions["from_cache"]
            assert response.extensions["cache_metadata"]["number_of_uses"] == 2
            stored = storage.retrieve(response.extensions["cache_metadata"]["cache_key"])
            assert stored
            stored_response, stored_request, stored_metadata = stored
            assert extract_header_values_decoded(stored_response.headers, b"Date") == ["Mon, 25 Aug 2015 12:00:11 GMT"]
            assert stored_response.content == b"Eat at Joe's."



def test_transport_revalidation_forward_extensions():
    class MockedClock(BaseClock):
        current = 1440504000  # Mon, 25 Aug 2015 12:00:00 GMT

        def now(self) -> int:
            return self.current

    class MockedTransportWithExtensionsMemory(hishel.MockTransport):
        def handle_request(self, request: httpx.Request) -> httpx.Response:
            self.last_request_extensions = request.extensions
            return super().handle_request(request)

    clock = MockedClock()
    controller = hishel.Controller(clock=clock)

    with MockedTransportWithExtensionsMemory() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    200,
                    headers=[
                        (b"Cache-Control", b"max-age=1"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:00 GMT"),
                    ],
                ),
                httpx.Response(
                    304,
                    headers=[
                        (b"Cache-Control", b"max-age=1"),
                        (b"Date", b"Mon, 25 Aug 2015 12:00:01 GMT"),
                    ],
                ),
            ]
        )
        with hishel.CacheTransport(
            transport=transport, controller=controller, storage=hishel.InMemoryStorage()
        ) as cache_transport:
            # first request with extensions
            cache_transport.handle_request(
                httpx.Request("GET", "https://www.example.com", extensions={"foo": "bar"})
            )
            assert transport.last_request_extensions["foo"] == "bar"

            # cache expires
            clock.current += 1

            # second request with extensions that should be passed to revalidation request
            response = cache_transport.handle_request(
                httpx.Request("GET", "https://www.example.com", extensions={"foo": "baz"})
            )
            assert response.extensions["revalidated"] is True
            assert transport.last_request_extensions["foo"] == "baz"



import httpx

from ._async import *
from ._controller import *
from ._exceptions import *
from ._headers import *
from ._serializers import *
from ._sync import *
from ._lfu_cache import *


def install_cache() -> None:  # pragma: no cover
    httpx.AsyncClient = AsyncCacheClient  # type: ignore
    httpx.Client = CacheClient  # type: ignore


__version__ = "0.1.2"



from __future__ import annotations

import datetime
import logging
import os
import time
import typing as t
import typing as tp
import warnings
from copy import deepcopy
from pathlib import Path

try:
    import boto3

    from .._s3 import S3Manager
except ImportError:  # pragma: no cover
    boto3 = None  # type: ignore

try:
    import sqlite3
except ImportError:  # pragma: no cover
    sqlite3 = None  # type: ignore

from httpcore import Request, Response

if t.TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import TypeAlias

from hishel._serializers import BaseSerializer, clone_model

from .._files import FileManager
from .._serializers import JSONSerializer, Metadata
from .._synchronization import Lock
from .._utils import float_seconds_to_int_milliseconds

logger = logging.getLogger("hishel.storages")

__all__ = (
    "BaseStorage",
    "FileStorage",
    "RedisStorage",
    "SQLiteStorage",
    "InMemoryStorage",
    "S3Storage",
)

StoredResponse: TypeAlias = tp.Tuple[Response, Request, Metadata]
RemoveTypes = tp.Union[str, Response]

try:
    import redis
except ImportError:  # pragma: no cover
    redis = None  # type: ignore


class BaseStorage:
    def __init__(
        self,
        serializer: tp.Optional[BaseSerializer] = None,
        ttl: tp.Optional[tp.Union[int, float]] = None,
    ) -> None:
        self._serializer = serializer or JSONSerializer()
        self._ttl = ttl

    def store(self, key: str, response: Response, request: Request, metadata: Metadata | None = None) -> None:
        raise NotImplementedError()

    def remove(self, key: RemoveTypes) -> None:
        raise NotImplementedError()

    def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
        raise NotImplementedError()

    def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()


class FileStorage(BaseStorage):
    """
    A simple file storage.

    :param serializer: Serializer capable of serializing and de-serializing http responses, defaults to None
    :type serializer: tp.Optional[BaseSerializer], optional
    :param base_path: A storage base path where the responses should be saved, defaults to None
    :type base_path: tp.Optional[Path], optional
    :param ttl: Specifies the maximum number of seconds that the response can be cached, defaults to None
    :type ttl: tp.Optional[tp.Union[int, float]], optional
    :param check_ttl_every: How often in seconds to check staleness of **all** cache files.
        Makes sense only with set `ttl`, defaults to 60
    :type check_ttl_every: tp.Union[int, float]
    """

    def __init__(
        self,
        serializer: tp.Optional[BaseSerializer] = None,
        base_path: tp.Optional[Path] = None,
        ttl: tp.Optional[tp.Union[int, float]] = None,
        check_ttl_every: tp.Union[int, float] = 60,
    ) -> None:
        super().__init__(serializer, ttl)

        self._base_path = Path(base_path) if base_path is not None else Path(".cache/hishel")
        self._gitignore_file = self._base_path / ".gitignore"

        if not self._base_path.is_dir():
            self._base_path.mkdir(parents=True)

        if not self._gitignore_file.is_file():
            with open(self._gitignore_file, "w", encoding="utf-8") as f:
                f.write("# Automatically created by Hishel\n*")

        self._file_manager = FileManager(is_binary=self._serializer.is_binary)
        self._lock = Lock()
        self._check_ttl_every = check_ttl_every
        self._last_cleaned = time.monotonic()

    def store(self, key: str, response: Response, request: Request, metadata: Metadata | None = None) -> None:
        """
        Stores the response in the cache.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additional information about the stored response
        :type metadata: Optional[Metadata]
        """

        metadata = metadata or Metadata(
            cache_key=key, created_at=datetime.datetime.now(datetime.timezone.utc), number_of_uses=0
        )
        response_path = self._base_path / key

        with self._lock:
            self._file_manager.write_to(
                str(response_path),
                self._serializer.dumps(response=response, request=request, metadata=metadata),
            )
        self._remove_expired_caches(response_path)

    def remove(self, key: RemoveTypes) -> None:
        """
        Removes the response from the cache.

        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
        :type key: Union[str, Response]
        """

        if isinstance(key, Response):  # pragma: no cover
            key = t.cast(str, key.extensions["cache_metadata"]["cache_key"])

        response_path = self._base_path / key

        with self._lock:
            if response_path.exists():
                response_path.unlink()

    def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
        """
        Updates the metadata of the stored response.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        """
        response_path = self._base_path / key

        with self._lock:
            if response_path.exists():
                atime = response_path.stat().st_atime
                old_mtime = response_path.stat().st_mtime
                self._file_manager.write_to(
                    str(response_path),
                    self._serializer.dumps(response=response, request=request, metadata=metadata),
                )

                # Restore the old atime and mtime (we use mtime to check the cache expiration time)
                os.utime(response_path, (atime, old_mtime))
                return

        return self.store(key, response, request, metadata)  # pragma: no cover

    def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        """
        Retreives the response from the cache using his key.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :return: An HTTP response and his HTTP request.
        :rtype: tp.Optional[StoredResponse]
        """

        response_path = self._base_path / key

        self._remove_expired_caches(response_path)
        with self._lock:
            if response_path.exists():
                read_data = self._file_manager.read_from(str(response_path))
                if len(read_data) != 0:
                    return self._serializer.loads(read_data)
        return None

    def close(self) -> None:  # pragma: no cover
        return

    def _remove_expired_caches(self, response_path: Path) -> None:
        if self._ttl is None:
            return

        if time.monotonic() - self._last_cleaned < self._check_ttl_every:
            if response_path.is_file():
                age = time.time() - response_path.stat().st_mtime
                if age > self._ttl:
                    response_path.unlink()
            return

        self._last_cleaned = time.monotonic()
        with self._lock:
            with os.scandir(self._base_path) as entries:
                for entry in entries:
                    try:
                        if entry.is_file():
                            age = time.time() - entry.stat().st_mtime
                            if age > self._ttl:
                                os.unlink(entry.path)
                    except FileNotFoundError:  # pragma: no cover
                        pass


class SQLiteStorage(BaseStorage):
    """
    A simple sqlite3 storage.

    :param serializer: Serializer capable of serializing and de-serializing http responses, defaults to None
    :type serializer: tp.Optional[BaseSerializer], optional
    :param connection: A connection for sqlite, defaults to None
    :type connection: tp.Optional[sqlite3.Connection], optional
    :param ttl: Specifies the maximum number of seconds that the response can be cached, defaults to None
    :type ttl: tp.Optional[tp.Union[int, float]], optional
    """

    def __init__(
        self,
        serializer: tp.Optional[BaseSerializer] = None,
        connection: tp.Optional[sqlite3.Connection] = None,
        ttl: tp.Optional[tp.Union[int, float]] = None,
    ) -> None:
        if sqlite3 is None:  # pragma: no cover
            raise RuntimeError(
                f"The `{type(self).__name__}` was used, but the required packages were not found. "
                "Check that you have `Hishel` installed with the `sqlite` extension as shown.\n"
                "```pip install hishel[sqlite]```"
            )
        super().__init__(serializer, ttl)

        self._connection: tp.Optional[sqlite3.Connection] = connection or None
        self._setup_lock = Lock()
        self._setup_completed: bool = False
        self._lock = Lock()

    def _setup(self) -> None:
        with self._setup_lock:
            if not self._setup_completed:
                if not self._connection:  # pragma: no cover
                    self._connection = sqlite3.connect(".hishel.sqlite", check_same_thread=False)
                self._connection.execute(
                    "CREATE TABLE IF NOT EXISTS cache(key TEXT, data BLOB, date_created REAL)"
                )
                self._connection.commit()
                self._setup_completed = True

    def store(self, key: str, response: Response, request: Request, metadata: Metadata | None = None) -> None:
        """
        Stores the response in the cache.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additioal information about the stored response
        :type metadata: Optional[Metadata]
        """

        self._setup()
        assert self._connection

        metadata = metadata or Metadata(
            cache_key=key, created_at=datetime.datetime.now(datetime.timezone.utc), number_of_uses=0
        )

        with self._lock:
            self._connection.execute("DELETE FROM cache WHERE key = ?", [key])
            serialized_response = self._serializer.dumps(response=response, request=request, metadata=metadata)
            self._connection.execute(
                "INSERT INTO cache(key, data, date_created) VALUES(?, ?, ?)", [key, serialized_response, time.time()]
            )
            self._connection.commit()
        self._remove_expired_caches()

    def remove(self, key: RemoveTypes) -> None:
        """
        Removes the response from the cache.

        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
        :type key: Union[str, Response]
        """

        self._setup()
        assert self._connection

        if isinstance(key, Response):  # pragma: no cover
            key = t.cast(str, key.extensions["cache_metadata"]["cache_key"])

        with self._lock:
            self._connection.execute("DELETE FROM cache WHERE key = ?", [key])
            self._connection.commit()

    def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
        """
        Updates the metadata of the stored response.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        """

        self._setup()
        assert self._connection

        with self._lock:
            cursor = self._connection.execute("SELECT data FROM cache WHERE key = ?", [key])
            row = cursor.fetchone()
            if row is not None:
                serialized_response = self._serializer.dumps(response=response, request=request, metadata=metadata)
                self._connection.execute("UPDATE cache SET data = ? WHERE key = ?", [serialized_response, key])
                self._connection.commit()
                return
        return self.store(key, response, request, metadata)  # pragma: no cover

    def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        """
        Retreives the response from the cache using his key.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :return: An HTTP response and its HTTP request.
        :rtype: tp.Optional[StoredResponse]
        """

        self._setup()
        assert self._connection

        self._remove_expired_caches()
        with self._lock:
            cursor = self._connection.execute("SELECT data FROM cache WHERE key = ?", [key])
            row = cursor.fetchone()
            if row is None:
                return None

            cached_response = row[0]
            return self._serializer.loads(cached_response)

    def close(self) -> None:  # pragma: no cover
        if self._connection is not None:
            self._connection.close()

    def _remove_expired_caches(self) -> None:
        assert self._connection
        if self._ttl is None:
            return

        with self._lock:
            self._connection.execute("DELETE FROM cache WHERE date_created + ? < ?", [self._ttl, time.time()])
            self._connection.commit()


class RedisStorage(BaseStorage):
    """
    A simple redis storage.

    :param serializer: Serializer capable of serializing and de-serializing http responses, defaults to None
    :type serializer: tp.Optional[BaseSerializer], optional
    :param client: A client for redis, defaults to None
    :type client: tp.Optional["redis.Redis"], optional
    :param ttl: Specifies the maximum number of seconds that the response can be cached, defaults to None
    :type ttl: tp.Optional[tp.Union[int, float]], optional
    """

    def __init__(
        self,
        serializer: tp.Optional[BaseSerializer] = None,
        client: tp.Optional[redis.Redis] = None,  # type: ignore
        ttl: tp.Optional[tp.Union[int, float]] = None,
    ) -> None:
        if redis is None:  # pragma: no cover
            raise RuntimeError(
                f"The `{type(self).__name__}` was used, but the required packages were not found. "
                "Check that you have `Hishel` installed with the `redis` extension as shown.\n"
                "```pip install hishel[redis]```"
            )
        super().__init__(serializer, ttl)

        if client is None:
            self._client = redis.Redis()  # type: ignore
        else:  # pragma: no cover
            self._client = client

    def store(self, key: str, response: Response, request: Request, metadata: Metadata | None = None) -> None:
        """
        Stores the response in the cache.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additioal information about the stored response
        :type metadata: Optional[Metadata]
        """

        metadata = metadata or Metadata(
            cache_key=key, created_at=datetime.datetime.now(datetime.timezone.utc), number_of_uses=0
        )

        if self._ttl is not None:
            px = float_seconds_to_int_milliseconds(self._ttl)
        else:
            px = None

        self._client.set(
            key, self._serializer.dumps(response=response, request=request, metadata=metadata), px=px
        )

    def remove(self, key: RemoveTypes) -> None:
        """
        Removes the response from the cache.

        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
        :type key: Union[str, Response]
        """

        if isinstance(key, Response):  # pragma: no cover
            key = t.cast(str, key.extensions["cache_metadata"]["cache_key"])

        self._client.delete(key)

    def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
        """
        Updates the metadata of the stored response.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        """

        ttl_in_milliseconds = self._client.pttl(key)

        # -2: if the key does not exist in Redis
        # -1: if the key exists in Redis but has no expiration
        if ttl_in_milliseconds == -2 or ttl_in_milliseconds == -1:  # pragma: no cover
            self.store(key, response, request, metadata)
        else:
            self._client.set(
                key,
                self._serializer.dumps(response=response, request=request, metadata=metadata),
                px=ttl_in_milliseconds,
            )

    def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        """
        Retreives the response from the cache using his key.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :return: An HTTP response and its HTTP request.
        :rtype: tp.Optional[StoredResponse]
        """

        cached_response = self._client.get(key)
        if cached_response is None:
            return None

        return self._serializer.loads(cached_response)

    def close(self) -> None:  # pragma: no cover
        self._client.close()


class InMemoryStorage(BaseStorage):
    """
    A simple in-memory storage.

    :param serializer: Serializer capable of serializing and de-serializing http responses, defaults to None
    :type serializer: tp.Optional[BaseSerializer], optional
    :param ttl: Specifies the maximum number of seconds that the response can be cached, defaults to None
    :type ttl: tp.Optional[tp.Union[int, float]], optional
    :param capacity: The maximum number of responses that can be cached, defaults to 128
    :type capacity: int, optional
    """

    def __init__(
        self,
        serializer: tp.Optional[BaseSerializer] = None,
        ttl: tp.Optional[tp.Union[int, float]] = None,
        capacity: int = 128,
    ) -> None:
        super().__init__(serializer, ttl)

        if serializer is not None:  # pragma: no cover
            warnings.warn("The serializer is not used in the in-memory storage.", RuntimeWarning)

        from hishel import LFUCache

        self._cache: LFUCache[str, tp.Tuple[StoredResponse, float]] = LFUCache(capacity=capacity)
        self._lock = Lock()

    def store(self, key: str, response: Response, request: Request, metadata: Metadata | None = None) -> None:
        """
        Stores the response in the cache.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additioal information about the stored response
        :type metadata: Optional[Metadata]
        """

        metadata = metadata or Metadata(
            cache_key=key, created_at=datetime.datetime.now(datetime.timezone.utc), number_of_uses=0
        )

        with self._lock:
            response_clone = clone_model(response)
            request_clone = clone_model(request)
            stored_response: StoredResponse = (deepcopy(response_clone), deepcopy(request_clone), metadata)
            self._cache.put(key, (stored_response, time.monotonic()))
        self._remove_expired_caches()

    def remove(self, key: RemoveTypes) -> None:
        """
        Removes the response from the cache.

        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
        :type key: Union[str, Response]
        """

        if isinstance(key, Response):  # pragma: no cover
            key = t.cast(str, key.extensions["cache_metadata"]["cache_key"])

        with self._lock:
            self._cache.remove_key(key)

    def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
        """
        Updates the metadata of the stored response.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        """

        with self._lock:
            try:
                stored_response, created_at = self._cache.get(key)
                stored_response = (stored_response[0], stored_response[1], metadata)
                self._cache.put(key, (stored_response, created_at))
                return
            except KeyError:  # pragma: no cover
                pass
        self.store(key, response, request, metadata)  # pragma: no cover

    def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        """
        Retreives the response from the cache using his key.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :return: An HTTP response and its HTTP request.
        :rtype: tp.Optional[StoredResponse]
        """

        self._remove_expired_caches()
        with self._lock:
            try:
                stored_response, _ = self._cache.get(key)
            except KeyError:
                return None
            return stored_response

    def close(self) -> None:  # pragma: no cover
        return

    def _remove_expired_caches(self) -> None:
        if self._ttl is None:
            return

        with self._lock:
            keys_to_remove = set()

            for key in self._cache:
                created_at = self._cache.get(key)[1]

                if time.monotonic() - created_at > self._ttl:
                    keys_to_remove.add(key)

            for key in keys_to_remove:
                self._cache.remove_key(key)


class S3Storage(BaseStorage):  # pragma: no cover
    """
    AWS S3 storage.

    :param bucket_name: The name of the bucket to store the responses in
    :type bucket_name: str
    :param serializer: Serializer capable of serializing and de-serializing http responses, defaults to None
    :type serializer: tp.Optional[BaseSerializer], optional
    :param ttl: Specifies the maximum number of seconds that the response can be cached, defaults to None
    :type ttl: tp.Optional[tp.Union[int, float]], optional
    :param check_ttl_every: How often in seconds to check staleness of **all** cache files.
        Makes sense only with set `ttl`, defaults to 60
    :type check_ttl_every: tp.Union[int, float]
    :param client: A client for S3, defaults to None
    :type client: tp.Optional[tp.Any], optional
    """

    def __init__(
        self,
        bucket_name: str,
        serializer: tp.Optional[BaseSerializer] = None,
        ttl: tp.Optional[tp.Union[int, float]] = None,
        check_ttl_every: tp.Union[int, float] = 60,
        client: tp.Optional[tp.Any] = None,
    ) -> None:
        super().__init__(serializer, ttl)

        if boto3 is None:  # pragma: no cover
            raise RuntimeError(
                f"The `{type(self).__name__}` was used, but the required packages were not found. "
                "Check that you have `Hishel` installed with the `s3` extension as shown.\n"
                "```pip install hishel[s3]```"
            )

        self._bucket_name = bucket_name
        client = client or boto3.client("s3")
        self._s3_manager = S3Manager(
            client=client,
            bucket_name=bucket_name,
            is_binary=self._serializer.is_binary,
            check_ttl_every=check_ttl_every,
        )
        self._lock = Lock()

    def store(self, key: str, response: Response, request: Request, metadata: Metadata | None = None) -> None:
        """
        Stores the response in the cache.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additioal information about the stored response
        :type metadata: Optional[Metadata]`
        """

        metadata = metadata or Metadata(
            cache_key=key, created_at=datetime.datetime.now(datetime.timezone.utc), number_of_uses=0
        )

        with self._lock:
            serialized = self._serializer.dumps(response=response, request=request, metadata=metadata)
            self._s3_manager.write_to(path=key, data=serialized)

        self._remove_expired_caches(key)

    def remove(self, key: RemoveTypes) -> None:
        """
        Removes the response from the cache.

        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
        :type key: Union[str, Response]
        """

        if isinstance(key, Response):  # pragma: no cover
            key = t.cast(str, key.extensions["cache_metadata"]["cache_key"])

        with self._lock:
            self._s3_manager.remove_entry(key)

    def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
        """
        Updates the metadata of the stored response.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :param response: An HTTP response
        :type response: httpcore.Response
        :param request: An HTTP request
        :type request: httpcore.Request
        :param metadata: Additional information about the stored response
        :type metadata: Metadata
        """

        with self._lock:
            serialized = self._serializer.dumps(response=response, request=request, metadata=metadata)
            self._s3_manager.write_to(path=key, data=serialized, only_metadata=True)

    def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        """
        Retreives the response from the cache using his key.

        :param key: Hashed value of concatenated HTTP method and URI
        :type key: str
        :return: An HTTP response and its HTTP request.
        :rtype: tp.Optional[StoredResponse]
        """

        self._remove_expired_caches(key)
        with self._lock:
            try:
                return self._serializer.loads(self._s3_manager.read_from(path=key))
            except Exception:
                return None

    def close(self) -> None:  # pragma: no cover
        return

    def _remove_expired_caches(self, key: str) -> None:
        if self._ttl is None:
            return

        with self._lock:
            converted_ttl = float_seconds_to_int_milliseconds(self._ttl)
            self._s3_manager.remove_expired(ttl=converted_ttl, key=key)



import typing as tp
from types import TracebackType

import httpcore
import httpx
from httpcore._sync.interfaces import RequestInterface

if tp.TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import Self

__all__ = ("MockConnectionPool", "MockTransport")


class MockConnectionPool(RequestInterface):
    def handle_request(self, request: httpcore.Request) -> httpcore.Response:
        assert isinstance(request.stream, tp.Iterable)
        data = b"".join([chunk for chunk in request.stream])  # noqa: F841
        return self.mocked_responses.pop(0)

    def add_responses(self, responses: tp.List[httpcore.Response]) -> None:
        if not hasattr(self, "mocked_responses"):
            self.mocked_responses = []
        self.mocked_responses.extend(responses)

    def __enter__(self) -> "Self":
        return self

    def __exit__(
        self,
        exc_type: tp.Optional[tp.Type[BaseException]] = None,
        exc_value: tp.Optional[BaseException] = None,
        traceback: tp.Optional[TracebackType] = None,
    ) -> None: ...


class MockTransport(httpx.BaseTransport):
    def handle_request(self, request: httpx.Request) -> httpx.Response:
        return self.mocked_responses.pop(0)

    def add_responses(self, responses: tp.List[httpx.Response]) -> None:
        if not hasattr(self, "mocked_responses"):
            self.mocked_responses = []
        self.mocked_responses.extend(responses)



import os
from datetime import datetime, timedelta
from pathlib import Path
from time import mktime
from wsgiref.handlers import format_date_time

import httpx
import pytest
from httpcore import Request

import hishel
from hishel._utils import generate_key

date_header = format_date_time(mktime((datetime.now() - timedelta(hours=2)).timetuple()))


@pytest.mark.anyio
async def test_client_301():
    async with hishel.MockAsyncTransport() as transport:
        transport.add_responses([httpx.Response(301, headers=[(b"Location", b"https://example.com")])])
        async with hishel.AsyncCacheClient(transport=transport, storage=hishel.AsyncInMemoryStorage()) as client:
            await client.request(
                "GET",
                "https://www.example.com",
            )
            response = await client.request(
                "GET",
                "https://www.example.com",
            )
            assert response.extensions["from_cache"]


@pytest.mark.anyio
async def test_empty_cachefile_handling(use_temp_dir):
    async with hishel.MockAsyncTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    status_code=200,
                    headers=[
                        ("Cache-Control", "public, max-age=86400, s-maxage=86400"),
                        ("Date", date_header),
                    ],
                    text="test",
                )
                for _ in range(2)
            ]
        )

        async with hishel.AsyncCacheClient(storage=hishel.AsyncFileStorage(), transport=transport) as client:
            request = Request(b"GET", "https://example.com/")
            key = generate_key(request)
            filedir = Path(os.getcwd() + "/.cache/hishel/" + key)

            await client.get("https://example.com/")
            response = await client.get("https://example.com/")

            assert response.status_code == 200
            assert response.text == "test"
            assert response.extensions["from_cache"]

            with open(filedir, "w+", encoding="utf-8") as file:
                file.truncate(0)
            assert os.path.getsize(filedir) == 0

            response = await client.get("https://example.com/")
            assert response.status_code == 200
            assert response.text == "test"
            assert response.extensions["from_cache"] is False

            response = await client.get("https://example.com/")
            assert response.extensions["from_cache"]


@pytest.mark.anyio
async def test_post_caching():
    async with hishel.MockAsyncTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    status_code=200,
                    headers=[
                        ("Cache-Control", "public, max-age=86400, s-maxage=86400"),
                        ("Date", date_header),
                    ],
                    text=f"test-{idx}",
                )
                for idx in range(2)
            ]
        )

        async with hishel.AsyncCacheClient(
            storage=hishel.AsyncInMemoryStorage(),
            transport=transport,
            controller=hishel.Controller(cacheable_methods=["POST"]),
        ) as client:
            # Create cache file.
            response = await client.post("https://example.com", json={"test": 1})
            assert response.status_code == 200
            assert not response.extensions["from_cache"]
            assert response.text == "test-0"

            # Get from cache file.
            response = await client.post("https://example.com", json={"test": 1})
            assert response.status_code == 200
            assert response.extensions["from_cache"]
            assert response.text == "test-0"

            # Create a new cache file
            response = await client.post("https://example.com", json={"test": 2})
            assert response.status_code == 200
            assert not response.extensions["from_cache"]
            assert response.text == "test-1"

            # Take second response from cache
            response = await client.post("https://example.com", json={"test": 2})
            assert response.status_code == 200
            assert response.extensions["from_cache"]
            assert response.text == "test-1"

            # Check on first response
            response = await client.post("https://example.com", json={"test": 1})
            assert response.status_code == 200
            assert response.extensions["from_cache"]
            assert response.text == "test-0"


@pytest.mark.anyio
async def test_client_get():
    async with hishel.MockAsyncTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    status_code=200,
                    headers=[
                        ("Cache-Control", "public, max-age=86400, s-maxage=86400"),
                        ("Date", date_header),
                    ],
                    text="test text",
                )
            ]
        )

        async with hishel.AsyncCacheClient(storage=hishel.AsyncInMemoryStorage(), transport=transport) as client:
            response = await client.get("https://example.com")
            assert response.status_code == 200
            assert not response.extensions["from_cache"]
            assert response.text == "test text"

            response = await client.get("https://example.com")
            assert response.status_code == 200
            assert response.extensions["from_cache"]
            assert response.text == "test text"


@pytest.mark.anyio
async def test_client_head():
    async with hishel.MockAsyncTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    status_code=200,
                    headers=[
                        ("Cache-Control", "public, max-age=86400, s-maxage=86400"),
                        ("Date", date_header),
                    ],
                )
                for _ in range(2)
            ]
        )

        async with hishel.AsyncCacheClient(storage=hishel.AsyncInMemoryStorage(), transport=transport) as client:
            response = await client.head("https://example.com")
            assert response.status_code == 200
            assert not response.extensions["from_cache"]

            response = await client.head("https://example.com")
            assert response.status_code == 200
            assert not response.extensions["from_cache"]


@pytest.mark.anyio
async def test_force_cache():
    async with hishel.MockAsyncTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    status_code=200,
                    headers=[
                        ("Cache-Control", "no-store"),
                        ("Date", date_header),
                    ],
                )
                for _ in range(3)
            ]
        )

        async with hishel.AsyncCacheClient(
            storage=hishel.AsyncInMemoryStorage(),
            controller=hishel.Controller(cacheable_methods=["HEAD"]),
            transport=transport,
        ) as client:
            response = await client.head("https://example.com")
            assert response.status_code == 200
            assert not response.extensions["from_cache"]

            # Check that "no-store" is respected
            response = await client.head("https://example.com")
            assert response.status_code == 200
            assert not response.extensions["from_cache"]

            response = await client.head("https://example.com", extensions={"force_cache": True})
            assert response.status_code == 200
            assert not response.extensions["from_cache"]

            response = await client.head("https://example.com", extensions={"force_cache": True})
            assert response.status_code == 200
            assert response.extensions["from_cache"]


@pytest.mark.anyio
async def test_cache_disabled():
    async with hishel.MockAsyncTransport() as transport:
        transport.add_responses(
            [
                httpx.Response(
                    status_code=200,
                    headers=[
                        ("Cache-Control", "public, max-age=86400, s-maxage=86400"),
                        ("Date", date_header),
                    ],
                )
                for _ in range(2)
            ]
        )

        async with hishel.AsyncCacheClient(storage=hishel.AsyncInMemoryStorage(), transport=transport) as client:
            response = await client.get(
                "https://www.example.com/cacheable-endpoint", extensions={"cache_disabled": True}
            )
            assert response.status_code == 200
            assert not response.extensions["from_cache"]

            response = await client.get(
                "https://www.example.com/cacheable-endpoint", extensions={"cache_disabled": True}
            )
            assert response.status_code == 200
            assert not response.extensions["from_cache"]



from collections import OrderedDict
from typing import DefaultDict, Dict, Generic, Iterator, Tuple, TypeVar

K = TypeVar("K")
V = TypeVar("V")

__all__ = ["LFUCache"]


class LFUCache(Generic[K, V]):
    def __init__(self, capacity: int):
        if capacity <= 0:
            raise ValueError("Capacity must be positive")

        self.capacity = capacity
        self.cache: Dict[K, Tuple[V, int]] = {}  # To store key-value pairs
        self.freq_count: DefaultDict[int, OrderedDict[K, V]] = DefaultDict(
            OrderedDict
        )  # To store frequency of each key
        self.min_freq = 0  # To keep track of the minimum frequency

    def get(self, key: K) -> V:
        if key in self.cache:
            value, freq = self.cache[key]
            # Update frequency and move the key to the next frequency bucket
            self.freq_count[freq].pop(key)
            if not self.freq_count[freq]:  # If the current frequency has no keys, update min_freq
                del self.freq_count[freq]
                if freq == self.min_freq:
                    self.min_freq += 1
            freq += 1
            self.freq_count[freq][key] = value
            self.cache[key] = (value, freq)
            return value
        raise KeyError(f"Key {key} not found")

    def put(self, key: K, value: V) -> None:
        if key in self.cache:
            _, freq = self.cache[key]
            # Update frequency and move the key to the next frequency bucket
            self.freq_count[freq].pop(key)
            if not self.freq_count[freq]:
                del self.freq_count[freq]
                if freq == self.min_freq:
                    self.min_freq += 1
            freq += 1
            self.freq_count[freq][key] = value
            self.cache[key] = (value, freq)
        else:
            # Check if cache is full, evict the least frequently used item
            if len(self.cache) == self.capacity:
                evicted_key, _ = self.freq_count[self.min_freq].popitem(last=False)
                del self.cache[evicted_key]

            # Add the new key-value pair with frequency 1
            self.cache[key] = (value, 1)
            self.freq_count[1][key] = value
            self.min_freq = 1

    def remove_key(self, key: K) -> None:
        if key in self.cache:
            _, freq = self.cache[key]
            self.freq_count[freq].pop(key)
            if not self.freq_count[freq]:  # If the current frequency has no keys, update min_freq
                del self.freq_count[freq]
                if freq == self.min_freq:
                    self.min_freq += 1
            del self.cache[key]

    def __iter__(self) -> Iterator[K]:
        yield from self.cache



import typing as tp

import anyio


class AsyncBaseFileManager:
    def __init__(self, is_binary: bool) -> None:
        self.is_binary = is_binary

    async def write_to(self, path: str, data: tp.Union[bytes, str], is_binary: tp.Optional[bool] = None) -> None:
        raise NotImplementedError()

    async def read_from(self, path: str, is_binary: tp.Optional[bool] = None) -> tp.Union[bytes, str]:
        raise NotImplementedError()


class AsyncFileManager(AsyncBaseFileManager):
    async def write_to(self, path: str, data: tp.Union[bytes, str], is_binary: tp.Optional[bool] = None) -> None:
        is_binary = self.is_binary if is_binary is None else is_binary
        mode = "wb" if is_binary else "wt"
        async with await anyio.open_file(path, mode) as f:  # type: ignore[call-overload]
            await f.write(data)

    async def read_from(self, path: str, is_binary: tp.Optional[bool] = None) -> tp.Union[bytes, str]:
        is_binary = self.is_binary if is_binary is None else is_binary
        mode = "rb" if is_binary else "rt"

        async with await anyio.open_file(path, mode) as f:  # type: ignore[call-overload]
            return tp.cast(tp.Union[bytes, str], await f.read())


class BaseFileManager:
    def __init__(self, is_binary: bool) -> None:
        self.is_binary = is_binary

    def write_to(self, path: str, data: tp.Union[bytes, str], is_binary: tp.Optional[bool] = None) -> None:
        raise NotImplementedError()

    def read_from(self, path: str, is_binary: tp.Optional[bool] = None) -> tp.Union[bytes, str]:
        raise NotImplementedError()


class FileManager(BaseFileManager):
    def write_to(self, path: str, data: tp.Union[bytes, str], is_binary: tp.Optional[bool] = None) -> None:
        is_binary = self.is_binary if is_binary is None else is_binary
        mode = "wb" if is_binary else "wt"
        with open(path, mode) as f:
            f.write(data)

    def read_from(self, path: str, is_binary: tp.Optional[bool] = None) -> tp.Union[bytes, str]:
        is_binary = self.is_binary if is_binary is None else is_binary
        mode = "rb" if is_binary else "rt"
        with open(path, mode) as f:
            return tp.cast(tp.Union[bytes, str], f.read())



import pytest

from hishel import ParseError, ValidationError, Vary
from hishel._headers import parse_cache_control


def test_blank_directive():
    header = [","]
    with pytest.raises(ParseError, match="The directive should not be left blank."):
        parse_cache_control(header)


def test_blank_directive_after_ows_stripping():
    header = [" ,"]
    with pytest.raises(ParseError, match="The directive should not contain only whitespaces."):
        parse_cache_control(header)


def test_invalid_key_symbol():
    header = ["\x12 ,"]
    with pytest.raises(
        ParseError,
        match=r"The character ''\\x12'' is not permitted in the directive name.",
    ):
        parse_cache_control(header)


def test_blank_directive_value():
    header = ["max-age= ,"]
    with pytest.raises(ParseError, match="The directive value cannot be left blank."):
        parse_cache_control(header)


def test_blank_invalid_quotes():
    header = ['max-age="123,']
    with pytest.raises(ParseError, match="Invalid quotes around the value."):
        parse_cache_control(header)


def test_invalid_symbol_in_unquoted():
    header = ["max-age=1\x123,"]
    with pytest.raises(
        ParseError,
        match=r"The character ''\\x12'' is not permitted for the unquoted values.",
    ):
        parse_cache_control(header)


def test_invalid_symbol_in_quoted():
    header = ['max-age="\x123",']
    with pytest.raises(
        ParseError,
        match=r"The character ''\\x12'' is not permitted for the quoted values.",
    ):
        parse_cache_control(header)


def test_time_field_without_value():
    header = ["max-age"]
    with pytest.raises(ValidationError, match="The directive 'max_age' necessitates a value."):
        parse_cache_control(header)


def test_time_field_with_quote():
    header = ['max-age="123"']
    with pytest.raises(
        ValidationError,
        match="The argument 'max_age' should be an integer, but a quote was found.",
    ):
        parse_cache_control(header)


def test_time_field_invalid_int():
    header = ["max-age=123t1"]
    with pytest.raises(
        ValidationError,
        match="The argument 'max_age' should be an integer, but got ''123t1''.",
    ):
        parse_cache_control(header)


def test_boolean_fields_with_value():
    header = ["no-store=1"]
    with pytest.raises(
        ValidationError,
        match="The directive 'no_store' should have no value, but it does.",
    ):
        parse_cache_control(header)


def test_list_value_empty():
    header = ['no-cache="," ']
    with pytest.raises(ValidationError, match="The list value must not be empty."):
        parse_cache_control(header)


def test_single_directive_parsing():
    header = ["max-age=3600"]
    cache_control = parse_cache_control(header)
    assert cache_control.max_age == 3600


def test_multiple_directives_parsing():
    header = ["max-age=3600", "s-maxage=3600"]
    cache_control = parse_cache_control(header)
    assert cache_control.max_age == 3600
    assert cache_control.s_maxage == 3600


def test_boolean_directives_parsing():
    header = ["no-store", "public"]
    cache_control = parse_cache_control(header)
    assert cache_control.no_store
    assert cache_control.public


def test_list_directives_parsing():
    header = ['no-cache="age, authorization"']
    cache_control = parse_cache_control(header)
    assert cache_control.no_cache == ["age", "authorization"]


def test_multiple_list_directives_parsing():
    header = ['no-cache="age, authorization"', 'private="age, authorization"']
    cache_control = parse_cache_control(header)
    assert cache_control.no_cache == ["age", "authorization"]
    assert cache_control.private == ["age", "authorization"]


def test_blank_list_directives():
    header = ["no-cache, private"]
    cache_control = parse_cache_control(header)
    assert cache_control.no_cache == True  # noqa: E712
    assert cache_control.private == True  # noqa: E712


def test_single_vary_header():
    header = ["Accept, Location"]
    vary = Vary.from_value(header)
    assert vary._values == ["Accept", "Location"]


def test_multiple_vary_headers():
    header = ["Accept", "Location"]
    vary = Vary.from_value(header)
    assert vary._values == ["Accept", "Location"]


def test_multiple_vary_headers_with_multiple_values():
    header = ["Accept, Location", "Transfer-Encoding"]
    vary = Vary.from_value(header)
    assert vary._values == ["Accept", "Location", "Transfer-Encoding"]



import types
import typing as tp
from threading import Lock as T_LOCK

import anyio


class AsyncLock:
    def __init__(self) -> None:
        self._lock = anyio.Lock()

    async def __aenter__(self) -> None:
        await self._lock.acquire()

    async def __aexit__(
        self,
        exc_type: tp.Optional[tp.Type[BaseException]] = None,
        exc_value: tp.Optional[BaseException] = None,
        traceback: tp.Optional[types.TracebackType] = None,
    ) -> None:
        self._lock.release()


class Lock:
    def __init__(self) -> None:
        self._lock = T_LOCK()

    def __enter__(self) -> None:
        self._lock.acquire()

    def __exit__(
        self,
        exc_type: tp.Optional[tp.Type[BaseException]] = None,
        exc_value: tp.Optional[BaseException] = None,
        traceback: tp.Optional[types.TracebackType] = None,
    ) -> None:
        self._lock.release()



from __future__ import annotations

import types
import typing as tp

import httpcore
import httpx
from httpx import AsyncByteStream, Request, Response
from httpx._exceptions import ConnectError

from hishel._utils import extract_header_values_decoded, normalized_url

from .._controller import Controller, allowed_stale
from .._headers import parse_cache_control
from .._serializers import JSONSerializer, Metadata
from ._storages import AsyncBaseStorage, AsyncFileStorage

if tp.TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import Self

__all__ = ("AsyncCacheTransport",)


async def fake_stream(content: bytes) -> tp.AsyncIterable[bytes]:
    yield content


def generate_504() -> Response:
    return Response(status_code=504)


class AsyncCacheStream(AsyncByteStream):
    def __init__(self, httpcore_stream: tp.AsyncIterable[bytes]):
        self._httpcore_stream = httpcore_stream

    async def __aiter__(self) -> tp.AsyncIterator[bytes]:
        async for part in self._httpcore_stream:
            yield part

    async def aclose(self) -> None:
        if hasattr(self._httpcore_stream, "aclose"):
            await self._httpcore_stream.aclose()


class AsyncCacheTransport(httpx.AsyncBaseTransport):
    """
    An HTTPX Transport that supports HTTP caching.

    :param transport: `Transport` that our class wraps in order to add an HTTP Cache layer on top of
    :type transport: httpx.AsyncBaseTransport
    :param storage: Storage that handles how the responses should be saved., defaults to None
    :type storage: tp.Optional[AsyncBaseStorage], optional
    :param controller: Controller that manages the cache behavior at the specification level, defaults to None
    :type controller: tp.Optional[Controller], optional
    """

    def __init__(
        self,
        transport: httpx.AsyncBaseTransport,
        storage: tp.Optional[AsyncBaseStorage] = None,
        controller: tp.Optional[Controller] = None,
    ) -> None:
        self._transport = transport

        self._storage = storage if storage is not None else AsyncFileStorage(serializer=JSONSerializer())

        if not isinstance(self._storage, AsyncBaseStorage):  # pragma: no cover
            raise TypeError(f"Expected subclass of `AsyncBaseStorage` but got `{storage.__class__.__name__}`")

        self._controller = controller if controller is not None else Controller()

    async def handle_async_request(self, request: Request) -> Response:
        """
        Handles HTTP requests while also implementing HTTP caching.

        :param request: An HTTP request
        :type request: httpx.Request
        :return: An HTTP response
        :rtype: httpx.Response
        """

        if request.extensions.get("cache_disabled", False):
            request.headers.update(
                [
                    ("Cache-Control", "no-store"),
                    ("Cache-Control", "no-cache"),
                    *[("cache-control", value) for value in request.headers.get_list("cache-control")],
                ]
            )

        if request.method not in ["GET", "HEAD"]:
            # If the HTTP method is, for example, POST,
            # we must also use the request data to generate the hash.
            body_for_key = await request.aread()
            request.stream = AsyncCacheStream(fake_stream(body_for_key))
        else:
            body_for_key = b""

        # Construct the HTTPCore request because Controllers and Storages work with HTTPCore requests.
        httpcore_request = httpcore.Request(
            method=request.method,
            url=httpcore.URL(
                scheme=request.url.raw_scheme,
                host=request.url.raw_host,
                port=request.url.port,
                target=request.url.raw_path,
            ),
            headers=request.headers.raw,
            content=request.stream,
            extensions=request.extensions,
        )

        key = self._controller._key_generator(httpcore_request, body_for_key)
        stored_data = await self._storage.retrieve(key)

        request_cache_control = parse_cache_control(
            extract_header_values_decoded(request.headers.raw, b"Cache-Control")
        )

        if request_cache_control.only_if_cached and not stored_data:
            return generate_504()

        if stored_data:
            # Try using the stored response if it was discovered.

            stored_response, stored_request, metadata = stored_data

            # Immediately read the stored response to avoid issues when trying to access the response body.
            stored_response.read()

            res = self._controller.construct_response_from_cache(
                request=httpcore_request,
                response=stored_response,
                original_request=stored_request,
            )

            if isinstance(res, httpcore.Response):
                # Simply use the response if the controller determines it is ready for use.
                return await self._create_hishel_response(
                    key=key,
                    response=res,
                    request=httpcore_request,
                    cached=True,
                    revalidated=False,
                    metadata=metadata,
                )

            if request_cache_control.only_if_cached:
                return generate_504()

            if isinstance(res, httpcore.Request):
                # Controller has determined that the response needs to be re-validated.
                assert isinstance(res.stream, tp.AsyncIterable)
                revalidation_request = Request(
                    method=res.method.decode(),
                    url=normalized_url(res.url),
                    headers=res.headers,
                    stream=AsyncCacheStream(res.stream),
                    extensions=res.extensions,
                )
                try:
                    revalidation_response = await self._transport.handle_async_request(revalidation_request)
                except ConnectError:
                    # If there is a connection error, we can use the stale response if allowed.
                    if self._controller._allow_stale and allowed_stale(response=stored_response):
                        return await self._create_hishel_response(
                            key=key,
                            response=stored_response,
                            request=httpcore_request,
                            cached=True,
                            revalidated=False,
                            metadata=metadata,
                        )
                    raise  # pragma: no cover
                assert isinstance(revalidation_response.stream, tp.AsyncIterable)
                httpcore_revalidation_response = httpcore.Response(
                    status=revalidation_response.status_code,
                    headers=revalidation_response.headers.raw,
                    content=AsyncCacheStream(revalidation_response.stream),
                    extensions=revalidation_response.extensions,
                )

                # Merge headers with the stale response.
                final_httpcore_response = self._controller.handle_validation_response(
                    old_response=stored_response,
                    new_response=httpcore_revalidation_response,
                )

                await final_httpcore_response.aread()
                await revalidation_response.aclose()

                assert isinstance(final_httpcore_response.stream, tp.AsyncIterable)

                # RFC 9111: 4.3.3. Handling a Validation Response
                # A 304 (Not Modified) response status code indicates that the stored response can be updated and
                # reused. A full response (i.e., one containing content) indicates that none of the stored responses
                # nominated in the conditional request are suitable. Instead, the cache MUST use the full response to
                # satisfy the request. The cache MAY store such a full response, subject to its constraints.
                if revalidation_response.status_code != 304 and self._controller.is_cachable(
                    request=httpcore_request, response=final_httpcore_response
                ):
                    await self._storage.store(key, response=final_httpcore_response, request=httpcore_request)

                return await self._create_hishel_response(
                    key=key,
                    response=final_httpcore_response,
                    request=httpcore_request,
                    cached=revalidation_response.status_code == 304,
                    revalidated=True,
                    metadata=metadata,
                )

        regular_response = await self._transport.handle_async_request(request)
        assert isinstance(regular_response.stream, tp.AsyncIterable)
        httpcore_regular_response = httpcore.Response(
            status=regular_response.status_code,
            headers=regular_response.headers.raw,
            content=AsyncCacheStream(regular_response.stream),
            extensions=regular_response.extensions,
        )
        await httpcore_regular_response.aread()
        await httpcore_regular_response.aclose()

        if self._controller.is_cachable(request=httpcore_request, response=httpcore_regular_response):
            await self._storage.store(
                key,
                response=httpcore_regular_response,
                request=httpcore_request,
            )

        return await self._create_hishel_response(
            key=key,
            response=httpcore_regular_response,
            request=httpcore_request,
            cached=False,
            revalidated=False,
        )

    async def _create_hishel_response(
        self,
        key: str,
        response: httpcore.Response,
        request: httpcore.Request,
        cached: bool,
        revalidated: bool,
        metadata: Metadata | None = None,
    ) -> Response:
        if cached:
            assert metadata
            metadata["number_of_uses"] += 1
            await self._storage.update_metadata(key=key, request=request, response=response, metadata=metadata)
            response.extensions["from_cache"] = True  # type: ignore[index]
            response.extensions["cache_metadata"] = metadata  # type: ignore[index]
        else:
            response.extensions["from_cache"] = False  # type: ignore[index]
        response.extensions["revalidated"] = revalidated  # type: ignore[index]
        return Response(
            status_code=response.status,
            headers=response.headers,
            stream=AsyncCacheStream(fake_stream(response.content)),
            extensions=response.extensions,
        )

    async def aclose(self) -> None:
        await self._storage.aclose()
        await self._transport.aclose()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: tp.Optional[tp.Type[BaseException]] = None,
        exc_value: tp.Optional[BaseException] = None,
        traceback: tp.Optional[types.TracebackType] = None,
    ) -> None:
        await self.aclose()
